# Copyright 2016 Leo Barring <leo.barring@protonmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from chirp import chirp_common, directory, memmap, \
                  bitwise, settings, errors, util
from struct import pack
import logging

LOG = logging.getLogger(__name__)

SUPPORT_NONSPLIT_DUPLEX_ONLY = False
SUPPORT_SPLIT_BUT_DEFAULT_TO_NONSPLIT_ALWAYS = True
UNAMBIGUOUS_CROSS_MODES_ONLY = True

# With this setting enabled, some CHIRP settings are stored in
# thought-to-be junk/padding data of the channel memories.
# Enabling this feature while using CHIRP with an actual radio
# and any effects thereof is entirely the responsibility of the user.
ENABLE_DANGEROUS_EXPERIMENTAL_FEATURES = False

MEM_FORMAT = """
// data fields are generally written 0xff if they are unset
struct {

#seekto 0x0000;
    struct {
        struct {
            // 0-3
            bbcd rx_freq[4];

            // 4-7
            bbcd tx_freq[4];

            // 8-9 A-B
            struct {
                u8 digital:1,
                   invert:1,
                   high:6;
                u8 low;
            } tone[2];
            // tx_squelch on 0, rx_squelch on 1

            // C
            // the duplex sign is not used for memories,
            // but is kept for interface consistency
            u8 duplex_sign:2,
               compander:1,
               txpower:1,
               modulation_width:1,
               txrx_reverse:1,
               bcl:2;

            // D
            u8 scrambler_type:3,
               use_scrambler:1,
               opt_signal:2,
               ptt_id_edge:2;

            // E-F
            // u8 _unknown_000E[2];
            %s
        } data[128];
        struct {
            // 0-5, alt 8-D
            char entry[6];

            // 6-8, alt E-F
            char _unknown_0806[2];
        } names[128];

#seekto 0x0c20;
        bit present[128];

#seekto 0x0c30;
        bit priority[128];
    } channel_memory;

#seekto 0x0c00;
    struct {
        // 0-3
        bbcd rx_freq[4];

        // 4-7
        bbcd tx_freq[4]; // actually offset, but kept for name consistency

        // 8
        struct {
            u8 digital:1,
               invert:1,
               high:6;
            u8 low;
        } tone[2]; // tx_squelch on 0, rx_squelch on 1

        // C
        u8 duplex_sign:2,
           compander:1,
           txpower:1,
           modulation_width:1,
           txrx_reverse:1,
           bcl:2;

        // D
        u8 scrambler_type:3,
           use_scrambler:1,
           opt_signal:2,
           ptt_id_edge:2;

        // E-F
        // u8 _unknown_0C0E[2];
        %s

    } vfo_data[2];

#seekto 0xc40;
    struct {
        // 0-5
        char model_string[6]; // typically PX888D, unknown if rw or ro

        // 6-7
        u8 _unknown_0C46[2];

        // 8-9
        struct {
            bbcd lower_freq[2];
            bbcd upper_freq[2];
        } band_limits[2];
    } model_information;

#seekto 0x0c50;
    char radio_information_string[16];

#seekto 0x0c60;
    struct {
        // 0
        u8 ptt_cancel_sq:1,
           dis_ptt_id:1,
           workmode_b:2,
           use_roger_beep:1,
           msk_reverse:1,
           workmode_a:2;

        // 1
        u8 backlight_color:2,
           backlight_mode:2,
           dual_single_watch:1,
           auto_keylock:1,
           scan_mode:2;

        // 2
        u8 rx_stun:1,
           tx_stun:1,
           boot_message_mode:2,
           battery_save:1,
           key_beep:1,
           voice_announce:2;

        // 3
        bbcd squelch_level;

        // 4
        bbcd tx_timeout;

        // 5
        u8 allow_keypad:1,
           relay_without_disable_tail:1,
           _unknown_0C65:1,
           call_channel_active:1,
           vox_gain:4;

        // 6
        bbcd vox_delay;

        // 7
        bbcd vfo_step;

        // 8
        bbcd ptt_id_type;

        // 9
        u8 keypad_lock:1,
            _unknown_0C69_1:1,
            side_button_hold_mode:2,
            dtmf_sidetone:1,
            _unknown_0C69_2:1,
            side_button_click_mode:2;

        // A
        u8 roger_beep:4,
           main_watch:1,
           _unknown_0C6A:3;

        // B
        u8 channel_a;

        // C
        u8 channel_b;

        // D
        u8 priority_channel;

        // E
        u8 wait_time;

        // F
        u8 _unknown_0C6F;

        // 0-7 on next block
        u8 _unknown_0C70[8];

        // 8-D on next block
        char boot_message[6];
    } opt_settings;

#seekto 0x0c80;
    struct {
        // these fields are used for all ptt id forms (msk/dtmf/5t)
        // (only one can be active and stored at a time)
        // and different constraints are applied depending
        // on the ptt id type
        u8 entry[7];
        u8 length;
    } ptt_id_data[2];
    // 0 is BOT, 1 is EOT

#seekto 0x0c90;
    struct {
        // 0
        u8 _unknown_0C90;

        // 1
        u8 _unknown_0C91_1:3,
           channel_stepping:1,
           unknown_0C91_2:1,
           receive_range:2,
           unknown_0C91_3:1;

        // 2-3
        u8 _unknown_0C92[2];

        // 4-7
        u8 vfo_freq[4];

        // 8-F and two more blocks
        struct {
            u8 entry[4];
        } memory[10];
    } fm_radio;

#seekto 0x0cc0;
    struct {
        char id_code[4];
        struct {
            char entry[4];
        } phone_book[9];
    } msk_settings;

#seekto 0x0cf0;
    struct {
        // 0-3
        bbcd rx_freq[4];

        // 4-7
        bbcd tx_freq[4];

        // 8
        struct {
            u8 digital:1,
               invert:1,
               high:6;
            u8 low;
        } tone[2];
        // tx_squelch on 0, rx_squelch on 1


        // C
        // the duplex sign is not used for the CALL,
        // channel but is kept for interface consistency
        u8 duplex_sign:2,
           compander:1,
           txpower:1,
           modulation_width:1,
           txrx_reverse:1,
           bcl:2;

        // D
        u8 scrambler_type:3,
           use_scrambler:1,
           opt_signal:2,
           ptt_id_edge:2;

        // E-F
        // u8 _unknown_0CFE[2];
        %s
    } call_channel;

#seekto 0x0d00;
    struct {

        // DTMF codes are stored as hex half-bytes,
        // 0-9 A-D are mapped straight
        // DTMF '*' is HEX E, DTMF '#' is HEX F

        // 0x0d00
        struct {
            u8 digit_length;       // 0x05 to 0x14 corresponding to 50-200ms
            u8 inter_digit_pause;  // same
            u8 first_digit_length; // same
            u8 first_digit_delay;  // 0x02 to 0x14 corresponding to 100-1000ms
        } timing;

#seekto 0x0d30;
        u8 _unknown_0D30[2]; // 0-1
        u8 group_code;       // 2
        u8 reset_time;       // 3
        u8 alert_transpond;  // 4
        u8 id_code[4];       // 5-8
        u8 _unknown_0D39[4]; // 9-C
        u8 id_code_length;   // D
        u8 _unknown_0d3e[2]; // E-F

// 0x0d40
        u8 tx_stun_code[4];
        u8 _unknown_0D44[4];
        u8 tx_stun_code_length;
        u8 cancel_tx_stun_code_length;
        u8 cancel_tx_stun_code[4];
        u8 _unknown_0D4E[2];

// 0x0d50
        u8 rxtx_stun_code[4];
        u8 _unknown_0D54[4];
        u8 rxtx_stun_code_length;
        u8 cancel_rxtx_stun_code_length;
        u8 cancel_rxtx_stun_code[4];
        u8 _unknown_0D5E[2];

// 0x0d60
        struct {
            u8 entry[5];
            u8 _unknown_0D65[3];
            u8 length;
            u8 _unknown_0D69[7];
        } phone_book[9];
    } dtmf_settings;

#seekto 0x0e00;
    struct {
        u8 delay;
        u8 _unknown_0E01[5];
        u8 alert_transpond;
        u8 reset_time;
        u8 tone_standard;
        u8 id_code[3];

#seekto 0x0e20;
        struct {
            u8 period;
            u8 group_code:4,
               repeat_code:4;
        } tone_settings[4];
        // the order is ZVEI1 ZVEI2 CCIR1 CCITT

#seekto 0x0e40;
        // 5-Tone tone standard frequency table
        // unknown use, changing the values does not seem to have
        // any effect on the produced sound, but the values are not
        // overwritten either.
        il16 tone_frequency_table[16];

// 0xe60
        u8 tx_stun_code[5];
        u8 _unknown_0E65[3];
        u8 tx_stun_code_length;
        u8 cancel_tx_stun_code_length;
        u8 cancel_tx_stun_code[5];
        u8 _unknown_0E6F;

// 0xe70
        u8 rxtx_stun_code[5];
        u8 _unknown_0E75[3];
        u8 rxtx_stun_code_length;
        u8 cancel_rxtx_stun_code_length;
        u8 cancel_rxtx_stun_code[5];
        u8 _unknown_0E7F;

// 0xe80
        struct {
            u8 entry[3];
        } phone_book[9];
    } five_tone_settings;
} mem;"""

# various magic numbers and strings, apart from the memory format
if ENABLE_DANGEROUS_EXPERIMENTAL_FEATURES:
    LOG.warn("ENABLE_DANGEROUS_EXPERIMENTAL_FEATURES"
             "AND/OR DANGEROUS FEATURES ENABLED")
    MEM_FORMAT = MEM_FORMAT % (
            "u8 _unknown_000E_1: 6,\n"
            "   experimental_duplex_mode_indicator: 1,\n"
            "   experimental_cross_mode_indicator: 1;\n"
            "u8 _unknown_000F;",
            "u8 _unknown_0C0E_1: 6,\n"
            "   experimental_duplex_mode_indicator: 1,\n"
            "   experimental_cross_mode_indicator: 1;\n"
            "u8 _unknown_0C0F;",
            "u8 _unknown_0CFE_1: 6,\n"
            "   experimental_duplex_mode_indicator: 1,\n"
            "   experimental_cross_mode_indicator: 1;\n"
            "u8 _unknown_0CFF;")
    # we don't need these settings anymore, because it's exactly what the
    # experimental features are about
    SUPPORT_SPLIT_BUT_DEFAULT_TO_NONSPLIT_ALWAYS = False
    SUPPORT_NONSPLIT_DUPLEX_ONLY = False
    UNAMBIGUOUS_CROSS_MODES_ONLY = False

else:
    MEM_FORMAT = MEM_FORMAT % (
            "u8 _unknown_000E[2];",
            "u8 _unknown_0C0E[2];",
            "u8 _unknown_0CFE[2];")

FILE_MAGIC = [0xc40, 0xc50,
              b'\x50\x58\x38\x38\x38\x44\x00\xff'
              b'\x13\x40\x17\x60\x40\x00\x48\x00']
HANDSHAKE_OUT = b'XONLINE'
HANDSHAKE_IN = [b'PX888D\x00\xff']

LOWER_READ_BOUND = 0
UPPER_READ_BOUND = 0x1000
LOWER_WRITE_BOUND = 0
UPPER_WRITE_BOUND = 0x0fc0
BLOCKSIZE = 64

OFF_INT = ["Off"] + [str(x+1) for x in range(100)]
OFF_ON = ["Off", "On"]
INACTIVE_ACTIVE = ["Inactive", "Active"]
NO_YES = ["No", "Yes"]
YES_NO = ["Yes", "No"]

BANDS = [(134000000, 176000000),  # VHF
         (400000000, 480000000)]  # UHF

SPECIAL_CHANNELS = {'VFO-A': -2, 'VFO-B': -1, 'CALL': 0}
SPECIAL_NUMBERS = {-2: 'VFO-A', -1: 'VFO-B', 0: 'CALL'}

DUPLEX_MODES = ['', '+', '-', 'split']
if SUPPORT_NONSPLIT_DUPLEX_ONLY:
    DUPLEX_MODES = ['', '+', '-']

TONE_MODES = ["", "Tone", "TSQL", "DTCS", "Cross"]

CROSS_MODES = ["Tone->Tone",
               "DTCS->",
               "->DTCS",
               "Tone->DTCS",
               "DTCS->Tone",
               "->Tone",
               "DTCS->DTCS",
               "Tone->"]
if UNAMBIGUOUS_CROSS_MODES_ONLY:
    CROSS_MODES = ["Tone->Tone",
                   "DTCS->",
                   "->DTCS",
                   "Tone->DTCS",
                   "DTCS->Tone",
                   "->Tone",
                   "DTCS->DTCS"]

MODES = ["NFM", "FM"]

# Only the 'High' power level is quantified in the manual for
# the radio (4W for VHF, 5W for UHF), a web search turned
# up numbers for the 'Low' power level (0.5W for VHF and
# 0.7W for UHF), but they are not official to my knowledge
# and should be taken with a grain of salt or two.
# Numbers used in code is the averages
POWER_LEVELS = [chirp_common.PowerLevel("Low", watts=0.6),
                chirp_common.PowerLevel("High", watts=4.5)]

SKIP_MODES = ["", "S"]
BCL_MODES = ["Off", "Carrier", "QT/DQT"]
SCRAMBLER_MODES = OFF_INT[0:9]
PTT_ID_EDGES = ["Off", "BOT", "EOT", "Both"]
OPTSIGN_MODES = ["None", "DTMF", "5-Tone", "MSK"]

VFO_STRIDE = ['5 kHz', '6.25 kHz', '10 kHz', '12.5 kHz', '25 kHz']
AB = ['A', 'B']
WATCH_MODES = ['Single watch', 'Dual watch']
AB_MODES = ['VFO', 'Memory index', 'Memory name', 'Memory frequency']
SCAN_MODES = ["Time", "Carrier", "Seek"]
WAIT_TIMES = [("0.3s", 6), ("0.5s", 10)] +\
             [("%ds" % t, t*20) for t in range(1, 13)]

BUTTON_MODES = ["Send call list data",
                "Emergency alarm",
                "Send 1750 Hz signal",
                "Open squelch"]
BOOT_MESSAGE_TYPES = ["Off", "Battery voltage", "Custom message"]
TALKBACK = ['Off', 'Chinese', 'English']
BACKLIGHT_COLORS = list(zip(["Blue", "Orange", "Purple"], range(1, 4)))
VOX_GAIN = OFF_INT[0:10]
VOX_DELAYS = ['1s', '2s', '3s', '4s']
TRANSMIT_ALARMS = ['Off', '30s', '60s', '90s', '120s',
                   '150s', '180s', '210s', '240s', '270s']

DATA_MODES = ['MSK', 'DTMF', '5-Tone']

ASCIIPART = ''.join([chr(x) for x in range(0x20, 0x7f)])
DTMF = "0123456789ABCD*#"
HEXADECIMAL = "0123456789ABCDEF"

ROGER_BEEP = OFF_INT[0:11]
BACKLIGHT_MODES = ["Off", "Auto", "On"]

TONE_RESET_TIME = ['Off'] + ['%ds' % x for x in range(1, 256)]
DTMF_TONE_RESET_TIME = TONE_RESET_TIME[0:16]

DTMF_GROUPS = list(zip(["Off", "A", "B", "C", "D", "*", "#"], [255] +
                       list(range(10, 16))))
FIVE_TONE_STANDARDS = ['ZVEI1', 'ZVEI2', 'CCIR1', 'CCITT']

# should mimic the defaults in the memedit MemoryEditor somewhat
#                          0   1   2   3   4   5   6   7
SANE_MEMORY_DEFAULT = b"\x14\x61\x00\x00\x14\x61\x00\x00" + \
                      b"\xff\xff\xff\xff\xc8\x00\xff\xff"
#                          8   9   A   B   C   D   E   F


# these two option sets are listed differently like this in the stock software,
# so I'm keeping them separate for now if they are in fact identical
# in behaviour, that should probably be amended
DTMF_ALERT_TRANSPOND = list(zip(['Off', 'Call alert',
                                 'Transpond-alert',
                                 'Transpond-ID code'],
                                [255] + list(range(1, 4))))
FIVE_TONE_ALERT_TRANSPOND = list(zip(['Off', 'Alert tone',
                                      'Transpond', 'Transpond-ID code'],
                                     [255] + list(range(1, 4))))

BFM_BANDS = ['87.5-108 MHz', '76.0-91.0 MHz', '76.0-108.0 MHz', '65.0-76.0 MHz']
BFM_STRIDE = ['100 kHz', '50 kHz']


def piperead(pipe, amount):
    """read some data, catch exceptions, validate length of data read"""
    try:
        d = pipe.read(amount)
    except Exception as e:
        raise errors.RadioError(
                "Tried to read %d bytes, but got an exception: %s" %
                (amount, repr(e)))
    if d is None:
        raise errors.RadioError(
                "Tried to read %d bytes, but read operation returned <None>." %
                (amount))
    if d is None or len(d) != amount:
        raise errors.RadioError(
                "Tried to read %d bytes, but got %d bytes instead." %
                (amount, len(d)))
    return d


def pipewrite(pipe, data):
    """write some data, catch exceptions, validate length of data written"""
    try:
        n = pipe.write(data)
    except Exception as e:
        raise errors.RadioError(
                "Tried to write %d bytes, but got an exception: %s." %
                (len(data), repr(e)))
    if n is None:
        raise errors.RadioError(
                "Tried to write %d bytes, but operation returned <None>." %
                (len(data)))
    if n != len(data):
        raise errors.RadioError(
                "Tried to write %d bytes, but wrote %d bytes instead." %
                (len(data), n))


def attempt_initial_handshake(pipe):
    """try to do the initial handshake"""
    pipewrite(pipe, HANDSHAKE_OUT)
    x = piperead(pipe, len(HANDSHAKE_IN[0]))
    if x in HANDSHAKE_IN:
        return True
    LOG.debug("Handshake failed: received: %s expected one of: %s" %
              (repr(x), repr(HANDSHAKE_IN)))
    return False


def initial_handshake(pipe, tries):
    """do an initial handshake attempt up to tries times"""
    x = False
    for i in range(tries):
        x = attempt_initial_handshake(pipe)
        if x:
            break
    if not x:
        raise errors.RadioError("Initial handshake failed all ten tries.")


def mk_writecommand(addr):
    """makes a write command from an address specification"""
    return pack('>cHc', b'W', addr, b'@')


def mk_readcommand(addr):
    """makes a read command from an address specification"""
    return pack('>cHc', b'R', addr, b'@')


def expect_ack(pipe):
    x = piperead(pipe, 1)
    if x != b'\x06':
        LOG.debug(
                "Did not get ACK. received: %s, expected: '\\x06'" %
                repr(x))
        raise errors.RadioError("Did not get ACK when expected.")


def end_communications(pipe):
    """tell the radio that we are done"""
    pipewrite(pipe, b'E')
    expect_ack(pipe)


def read_block(pipe, addr):
    """read and return a chunk of data at specified address"""
    r = mk_readcommand(addr)
    w = mk_writecommand(addr)
    pipewrite(pipe, r)
    x = piperead(pipe, len(w))
    if x != w:
        raise errors.RadioError("Received data not following protocol.")
    block = piperead(pipe, BLOCKSIZE)
    return block


def write_block(pipe, addr, block):
    """write a chunk of data at specified address"""
    w = mk_writecommand(addr)
    pipewrite(pipe, w)
    pipewrite(pipe, block)
    expect_ack(pipe)


def show_progress(radio, blockaddr, upper, msg):
    """relay read/write information to the user through the gui"""
    if radio.status_fn:
        status = chirp_common.Status()
        status.cur = blockaddr
        status.max = upper
        status.msg = msg
        radio.status_fn(status)


def do_download(radio):
    """download from the radio to the memory map"""
    initial_handshake(radio.pipe, 10)
    memory = memmap.MemoryMapBytes(b'\xff'*0x1000)
    for blockaddr in range(LOWER_READ_BOUND, UPPER_READ_BOUND, BLOCKSIZE):
        LOG.debug("Reading block "+str(blockaddr))
        block = read_block(radio.pipe, blockaddr)
        memory.set(blockaddr, block)
        show_progress(radio, blockaddr, UPPER_READ_BOUND,
                      "Reading radio memory... %04x" % blockaddr)
    end_communications(radio.pipe)
    return memory


def do_upload(radio):
    """upload from the memory map to the radio"""
    memory = radio.get_mmap()
    initial_handshake(radio.pipe, 10)
    for blockaddr in range(LOWER_WRITE_BOUND, UPPER_WRITE_BOUND, BLOCKSIZE):
        LOG.debug("Writing block "+str(blockaddr))
        block = memory[blockaddr:blockaddr+BLOCKSIZE]
        write_block(radio.pipe, blockaddr, block)
        show_progress(radio, blockaddr, UPPER_WRITE_BOUND,
                      "Writing radio memory...  % 04x" % blockaddr)
    end_communications(radio.pipe)


def parse_tone(t):
    """
    parse the tone (ctss, dtcs) part of the mmap
    into more easily handled data types
    """
    # [ mode, value, polarity ]
    if int(t.high) == 0x3f and int(t.low) == 0xff:
        return [None, None, None]
    elif bool(t.digital):
        t = ['DTCS',
             (int(t.high) & 0x0f)*100 +
             ((int(t.low) & 0xf0) >> 4)*10 +
             (int(t.low) & 0x0f),
             ['N', 'R'][bool(t.invert)]]
        if t[1] not in chirp_common.DTCS_CODES:
            return [None, None, None]
    else:
        t = ['Tone',
             ((int(t.high) & 0xf0) >> 4)*100 +
             (int(t.high) & 0x0f)*10 +
             ((int(t.low) & 0xf0) >> 4) +
             (int(t.low) & 0x0f)/10.0,
             None]
        if t[1] not in chirp_common.TONES:
            return [None, None, None]
    return t


def unparse_tone(t):
    """parse tone data back into the format used by the radio"""
    # [ mode, value, polarity ]
    if t[0] == 'Tone':
        tint = int(t[1]*10)
        t0, tint = tint % 10, tint // 10
        t1, tint = tint % 10, tint // 10
        t2, tint = tint % 10, tint // 10
        high = (tint << 4) | t2
        low = (t1 << 4) | t0
        digital = False
        invert = False
        return digital, invert, high, low
    elif t[0] == 'DTCS':
        tint = int(t[1])
        t0, tint = tint % 10, tint // 10
        t1, tint = tint % 10, tint // 10
        high = tint
        low = (t1 << 4) | t0
        digital = True
        invert = t[2] == 'R'
        return digital, invert, high, low
    return None


def decode_halfbytes(data, mapping, length):
    """
    construct a string from a datatype
    where each half-byte maps to a character
    """
    s = ''
    for i in range(length):
        if i & 1 == 0:
            s += mapping[(int(data[i >> 1]) & 0xf0) >> 4]
        else:
            s += mapping[int(data[i >> 1]) & 0x0f]
    return s


def encode_halfbytes(data, datapad, mapping, fillvalue, fieldlen):
    """encode data from a string where each character maps to a half-byte"""
    if len(data) & 1:
        # pad to an even length
        data += datapad
    o = [fillvalue] * fieldlen
    for i in range(0, len(data), 2):
        v = (mapping.index(data[i]) << 4) | mapping.index(data[i+1])
        o[i >> 1] = v
    return bytearray(util.byte_to_int(b) for b in o)


def decode_ffstring(data):
    """decode a string delimited by 0xff"""
    s = ''
    for b in data:
        if int(b) == 0xff:
            break
        s += chr(int(b))
    return s


def encode_ffstring(data, fieldlen):
    """right-pad to specified length with 0xff bytes"""
    extra = fieldlen-len(data)
    if extra > 0:
        data += '\xff'*extra
    return bytearray([util.byte_to_int(x) for x in data])


def decode_dtmf(data, length):
    """decode a field containing dtmf data into a string"""
    if length == 0xff:
        return ''
    return decode_halfbytes(data, DTMF, length)


def encode_dtmf(data, length, fieldlen):
    """encode a string containing dtmf characters into a data field"""
    return encode_halfbytes(data, '0', DTMF, b'\xff', fieldlen)


def decode_5tone(data):
    """decode a field containing 5-tone data into a string"""
    if (int(data[2]) & 0x0f) != 0:
        return ''
    return decode_halfbytes(data, HEXADECIMAL, 5)


def encode_5tone(data, fieldlen):
    """encode a string containing 5-tone characters into a data field"""
    return encode_halfbytes(data, '0', HEXADECIMAL, b'\xff', fieldlen)


def decode_freq(data):
    """decode frequency data for the broadcast FM radio memories"""
    data_out = ''
    if data[0] != 0xff:
        data_out = chirp_common.format_freq(
                int(decode_halfbytes(data, "0123456789", len(data)))*100000)
    return data_out


def encode_freq(data, fieldlen):
    """encode frequency data for the broadcast FM radio memories"""
    data_out = bytearray(b'\xff')*fieldlen
    if data != '':
        data_out = encode_halfbytes((('%%0%di' % (fieldlen << 1)) %
                                     int(chirp_common.parse_freq(data)/10)),
                                    '', '0123456789', '', fieldlen)
    return data_out


def sbyn(s, n):
    """setting by name"""
    return list(filter(lambda x: x.get_name() == n, s))[0]


# These helper classes provide a direct link between the value
# of the widget shown in the ui, and the setting in the memory
# map of the radio, lessening the need to write large chunks
# of code, first for populating the ui from the memory map,
# then secondly for parsing the values back.
# By supplying the memory map entry to the setting instance,
# it is possible to automatically 1) initialize the value of
# the setting, as well as 2) automatically update the memory
# value when the user changes it in the ui, without adding
# any code outside the class.
class MappedIntegerSettingValue(settings.RadioSettingValueInteger):
    """"
    Integer setting, with the possibility to add translation
    functions between memory map <-> integer setting
    """
    def __init__(self, val_mem, minval, maxval, step=1,
                 int_from_mem=lambda x: int(x),
                 mem_from_int=lambda x: x,
                 autowrite=True):
        """
        val_mem      - memory map entry for the value
        minval       - the minimum value allowed
        maxval       - maximum value allowed
        step         - value stepping
        int_from_mem - function to convert memory entry to integer
        mem_from_int - function to convert integer to memory entry
        autowrite    - automatically write the memory map entry
                       when the value is changed
        """
        self._val_mem = val_mem
        self._int_from_mem = int_from_mem
        self._mem_from_int = mem_from_int
        self._autowrite = autowrite
        settings.RadioSettingValueInteger.__init__(
                self,
                minval, maxval, self._int_from_mem(val_mem), step)

    def set_value(self, x):
        settings.RadioSettingValueInteger.set_value(self, x)
        if self._autowrite:
            self.write_mem()

    def write_mem(self):
        if self.get_mutable() and self._mem_from_int is not None:
            self._val_mem.set_value(self._mem_from_int(
                settings.RadioSettingValueInteger.get_value(self)))


class MappedListSettingValue(settings.RadioSettingValueMap):
    """Mapped list setting"""
    def __init__(self, val_mem, options, autowrite=True):
        """
        val_mem      - memory map entry for the value
        options      - either a list of strings options to present,
                       mapped to integers 0...n
                       in the memory map entry, or a list of tuples
                       ("option description", memory map value)
        int_from_mem - function to convert memory entry to integer
        mem_from_int - function to convert integer to memory entry
        autowrite    - automatically write the memory map entry when
                       the value is changed
        """
        self._val_mem = val_mem
        self._autowrite = autowrite
        options = list(options)
        if not isinstance(options[0], tuple):
            options = list(zip(options, range(len(options))))
        settings.RadioSettingValueMap.__init__(
                self,
                options, mem_val=int(val_mem))

    def set_value(self, value):
        settings.RadioSettingValueMap.set_value(self, value)
        if self._autowrite:
            self.write_mem()

    def write_mem(self):
        if self.get_mutable():
            self._val_mem.set_value(
                    settings.RadioSettingValueMap.get_mem_val(self))


class MappedCodedStringSettingValue(settings.RadioSettingValueString):
    """
    generic base class for a number of mapped presented-as-strings
    values which may need conversion between mem and string,
    and may store a length value in a separate mem field
    """
    def __init__(self, val_mem, len_mem, min_length, max_length,
                 charset=ASCIIPART, padchar=' ', autowrite=True,
                 str_from_mem=lambda mve, lve: str(mve[0:int(lve)]),
                 mem_val_from_str=lambda s, fl: s[0:fl],
                 mem_len_from_int=lambda l: l):
        """
        val_mem          - memory map entry for the value
        len_mem          - memory map entry for the length (or None)
        min_length       - length that the string will be right-padded to
        max_length       - maximum length of the string, set as maxlength
                           for the RadioSettingValueString
        charset          - the allowed charset
        padchar          - the character that will be used to pad short
                           strings, if not in the charset, charset[0] is used
        autowrite        - automatically call write_mem when the ui value
                           change
        str_from_mem     - function to convert from memory entry to string
                           value, form:
                           func(value_entry, length_entry or none) -> string
        mem_val_from_str - function to convert from string value to
                           memory-fitting value, form:
                           func(string, value_entry_length) ->
                           value to store in value entry
        mem_len_from_int - function to convert from string length to
                           memory-fitting value, form:
                           func(stringlength) -> value to store in length entry
        """
        self._min_length = min_length
        self._val_mem = val_mem
        self._len_mem = len_mem
        self._padchar = padchar
        if padchar not in charset:
            self._padchar = charset[0]
        self._autowrite = autowrite
        self._str_from_mem = str_from_mem
        self._mem_val_from_str = mem_val_from_str
        self._mem_len_from_int = mem_len_from_int
        settings.RadioSettingValueString.__init__(
                self,
                0, max_length,
                self._str_from_mem(self._val_mem, self._len_mem),
                charset=charset, autopad=False)

    def set_value(self, value):
        """
        Set the value of the string, pad if below minimum length,
        unless it's '' to provide a distinction between
        uninitialized/reset data and needs-to-be-padded data
        """
        while len(value) < self._min_length and len(value) != 0:
            value += self._padchar
        settings.RadioSettingValueString.set_value(self, value)
        if self._autowrite:
            self.write_mem()

    def write_mem(self):
        """update the memory"""
        if not self.get_mutable() or self._mem_val_from_str is None:
            return
        v = self.get_value()
        l = len(v)
        self._val_mem.set_value(self._mem_val_from_str(v, len(self._val_mem)))
        if self._len_mem is not None and self._mem_len_from_int is not None:
            self._len_mem.set_value(self._mem_len_from_int(l))


class MappedFFStringSettingValue(MappedCodedStringSettingValue):
    """
    Mapped string setting, tailored for the puxing px888k,
    which uses 0xff terminated strings.
    """
    def __init__(self, val_mem, min_length, max_length,
                 charset=ASCIIPART, padchar=' ', autowrite=True):
        MappedCodedStringSettingValue.__init__(
                self,
                val_mem, None, min_length, max_length,
                charset=charset, padchar=padchar, autowrite=autowrite,
                str_from_mem=lambda mve, lve: decode_ffstring(mve),
                mem_val_from_str=lambda s, fl: encode_ffstring(s, fl),
                mem_len_from_int=None)


class MappedDTMFStringSettingValue(MappedCodedStringSettingValue):
    """
    Mapped string setting, tailored for the puxing px888k
    field pairs (value and length) storing DTMF codes
    """
    def __init__(self, val_mem, len_mem, min_length, max_length,
                 autowrite=True):
        MappedCodedStringSettingValue.__init__(
                self,
                val_mem, len_mem, min_length, max_length,
                charset=DTMF, padchar='0', autowrite=autowrite,
                str_from_mem=lambda mve, lve: decode_dtmf(mve, lve),
                mem_val_from_str=lambda s, fl: encode_dtmf(s, len(s), fl))


class MappedFiveToneStringSettingValue(MappedCodedStringSettingValue):
    """
    Mapped string setting, tailored for the puxing px888k
    fields storing 5-Tone codes
    """
    def __init__(self, val_mem, autowrite=True):
        MappedCodedStringSettingValue.__init__(
                self,
                val_mem, None, 0, 5, charset=HEXADECIMAL, padchar='0',
                autowrite=autowrite,
                str_from_mem=lambda mve, lve: decode_5tone(mve),
                mem_val_from_str=lambda s, fl: encode_5tone(s, fl),
                mem_len_from_int=None)


class MappedFreqStringSettingValue(MappedCodedStringSettingValue):
    """
    Mapped string setting, tailored for the puxing px888k
    fields for the broadcast FM radio frequencies
    """
    def __init__(self, val_mem, autowrite=True):
        MappedCodedStringSettingValue.__init__(
                self,
                val_mem, None, 0, 128, charset=ASCIIPART, padchar=' ',
                autowrite=autowrite,
                str_from_mem=lambda mve, lve: decode_freq(mve),
                mem_val_from_str=lambda s, fl: encode_freq(s, fl))


# These functions lessen the amount of boilerplate on the form
# x = RadioSetting("AAA", "BBB", SomeKindOfRadioSettingValue( ... ))
# to
# x = some_kind_of_setting("AAA", "BBB", ... )
def integer_setting(k, n, *args, **kwargs):
    return settings.RadioSetting(
            k, n,
            MappedIntegerSettingValue(*args, **kwargs))


def list_setting(k, n, *args, **kwargs):
    return settings.RadioSetting(
            k, n,
            MappedListSettingValue(*args, **kwargs))


def ff_string_setting(k, n, *args, **kwargs):
    return settings.RadioSetting(
            k, n,
            MappedFFStringSettingValue(*args, **kwargs))


def dtmf_string_setting(k, n, *args, **kwargs):
    return settings.RadioSetting(
            k, n,
            MappedDTMFStringSettingValue(*args, **kwargs))


def five_tone_string_setting(k, n, *args, **kwargs):
    return settings.RadioSetting(
            k, n,
            MappedFiveToneStringSettingValue(*args, **kwargs))


def frequency_setting(k, n, *args, **kwargs):
    return settings.RadioSetting(
            k, n,
            MappedFreqStringSettingValue(*args, **kwargs))


@directory.register
class Puxing_PX888K_Radio(chirp_common.CloneModeRadio):
    """Puxing PX-888K"""
    VENDOR = "Puxing"
    MODEL = "PX-888K"
    BAUD_RATE = 9600
    NEEDS_COMPAT_SERIAL = False

    @classmethod
    def match_model(cls, filedata, filename):
        if len(filedata) == UPPER_READ_BOUND:
            if filedata[FILE_MAGIC[0]:FILE_MAGIC[1]] == FILE_MAGIC[2]:
                return True
        return False

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_bank_index = False
        rf.has_dtcs = True
        rf.has_ctone = True
        rf.has_rx_dtcs = True
        rf.has_dtcs_polarity = True
        rf.has_mode = True
        rf.has_offset = True
        rf.has_name = True
        rf.has_bank = False
        rf.has_bank_names = False
        rf.has_tuning_step = False
        rf.has_cross = True
        rf.has_infinite_number = False
        rf.has_nostep_tuning = False
        rf.has_comment = False
        rf.has_settings = True
        if SUPPORT_NONSPLIT_DUPLEX_ONLY:
            rf.can_odd_split = False
        else:
            rf.can_odd_split = True

        rf.valid_modes = MODES
        rf.valid_tmodes = TONE_MODES
        rf.valid_duplexes = DUPLEX_MODES
        rf.valid_bands = BANDS
        rf.valid_skips = SKIP_MODES
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_characters = ASCIIPART
        rf.valid_name_length = 6
        rf.valid_cross_modes = CROSS_MODES
        rf.memory_bounds = (1, 128)
        rf.valid_special_chans = list(SPECIAL_CHANNELS.keys())
        rf.valid_tuning_steps = [5.0, 6.25, 10.0, 12.5, 25.0]
        return rf

    def sync_in(self):
        self._mmap = do_download(self)
        self.process_mmap()

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def sync_out(self):
        do_upload(self)

    def _set_sane_defaults(self, data):
        # thanks thayward!
        data.set_raw(SANE_MEMORY_DEFAULT)

    def _uninitialize(self, data, n):
        if isinstance(data, bitwise.arrayDataElement):
            data.set_value(b"\xff"*n)
        else:
            data.set_raw(b"\xff"*n)

    def _get_memory_structs(self, number):
        """
        fetch the correct data structs no matter
        if its regular or special channels,
        no matter if they're referred by name or channel index
        """
        index = 2501
        i = -42
        designator = 'INVALID'
        isregular = False
        iscall = False
        isvfo = False
        _data = None
        _name = None
        _present = None
        _priority = None
        if number in SPECIAL_NUMBERS.keys():
            index = number
            # special by index
            designator = SPECIAL_NUMBERS[number]
        elif number in SPECIAL_CHANNELS.keys():
            # special by name
            index = SPECIAL_CHANNELS[number]
            designator = number
        elif number > 0:
            # regular by number
            index = number
            designator = number

        if index < 0:
            isvfo = True
            _data = self._memobj.mem.vfo_data[index+2]
        elif index == 0:
            iscall = True
            _data = self._memobj.mem.call_channel
        elif index > 0:
            isregular = True
            i = number - 1
            _data = self._memobj.mem.channel_memory.data[i]
            _name = self._memobj.mem.channel_memory.names[i].entry
            _present = self._memobj.mem.channel_memory.present[
                    (i & 0x78) | (7-(i & 0x07))]
            _priority = self._memobj.mem.channel_memory.priority[
                    (i & 0x78) | (7-(i & 0x07))]

        if _data == bytearray(0xff)*16:
            self._set_sane_defaults(_data)

        return (index, designator,
                _data, _name, _present, _priority,
                isregular, isvfo, iscall)

    def get_raw_memory(self, number):
        x = self._get_memory_structs(number)
        return repr(x[2])

    def get_memory(self, number):
        mem = chirp_common.Memory()
        (index, designator,
            _data, _name, _present, _priority,
            isregular, isvfo, iscall) = self._get_memory_structs(number)

        mem.number = index

        # handle empty channels
        if isregular:
            if bool(_present):
                mem.empty = False
                mem.name = str(decode_ffstring(_name))
                mem.skip = SKIP_MODES[1-int(_priority)]
            else:
                mem.empty = True
                mem.name = ''
                return mem
        else:
            mem.empty = False
            mem.name = ''
            mem.extd_number = designator

        # get frequency data
        mem.freq = int(_data.rx_freq)*10
        mem.offset = int(_data.tx_freq)*10

        # interpret frequency data
        # only the vfo channels support duplex,
        # memory channels operate in split mode all the time
        if isvfo:
            mem.duplex = DUPLEX_MODES[int(_data.duplex_sign)]
            if mem.duplex == '-':
                mem.offset = mem.freq - mem.offset
            elif mem.duplex == '':
                mem.offset = 0
            elif mem.duplex == '+':
                mem.offset = mem.offset - mem.freq
        else:
            if mem.freq == mem.offset:
                mem.duplex = ''
                mem.offset = 0
            elif SUPPORT_NONSPLIT_DUPLEX_ONLY or \
                    SUPPORT_SPLIT_BUT_DEFAULT_TO_NONSPLIT_ALWAYS:
                if mem.freq > mem.offset:
                    mem.offset = mem.freq - mem.offset
                    mem.duplex = '-'
                elif mem.freq < mem.offset:
                    mem.offset = mem.offset - mem.freq
                    mem.duplex = '+'
            else:
                mem.duplex = 'split'

        # get tone data
        txtone = parse_tone(_data.tone[0])
        rxtone = parse_tone(_data.tone[1])
        chirp_common.split_tone_decode(mem, txtone, rxtone)

######################################################################
        if ENABLE_DANGEROUS_EXPERIMENTAL_FEATURES:
            # override certain settings based on flags
            # that we have set in junk areas of the memory
            # or basically, we BELIEVE this to be junk memory,
            # hence why it's experimental and dangerous
            if bool(_data.experimental_cross_mode_indicator) is False:
                if mem.tmode == 'Tone':
                    mem.cross_mode = 'Tone->'
                elif mem.tmode == 'TSQL':
                    mem.cross_mode = 'Tone->Tone'
                elif mem.tmode == 'DTCS':
                    mem.cross_mode = 'DTCS->DTCS'
                mem.tmode = 'Cross'
######################################################################

        # transmit mode and power level
        mem.mode = MODES[bool(_data.modulation_width)]
        mem.power = POWER_LEVELS[_data.txpower]

        # extra channel settings
        mem.extra = settings.RadioSettingGroup(
                "extra",
                "extra",
                list_setting("Busy channel lockout",
                             "BCL",
                             _data.bcl,
                             BCL_MODES),
                list_setting("Swap transmit and receive frequencies",
                             "Tx Rx freq swap",
                             _data.txrx_reverse,
                             OFF_ON),
                list_setting("Use compander",
                             "Use compander",
                             _data.compander,
                             OFF_ON),
                list_setting("Use scrambler", "Use scrambler",
                             _data.use_scrambler,
                             NO_YES),
                list_setting("Scrambler selection",
                             "Voice Scrambler",
                             _data.scrambler_type,
                             SCRAMBLER_MODES),
                list_setting("Send ID code before and/or after transmitting",
                             "PTT ID",
                             _data.ptt_id_edge,
                             PTT_ID_EDGES),
                list_setting("Optional signal before/after transmission, " +
                             "this setting overrides the PTT ID setting",
                             "Opt Signal",
                             _data.opt_signal,
                             OPTSIGN_MODES))

######################################################################
        if ENABLE_DANGEROUS_EXPERIMENTAL_FEATURES:
            # override certain settings based on flags
            # that we have set in junk areas of the memory
            # or basically, we BELIEVE this to be junk memory,
            # hence why it's experimental and dangerous
            if bool(_data.experimental_duplex_mode_indicator) is False:
                # if this flag is set, this means that we in the gui
                # have set the duplex mode to something
                # the channel does not really support,
                # such as split modes for vfo channels,
                # and non-split modes for the memory channels
                mem.duplex = DUPLEX_MODES[int(_data.duplex_sign)]
                mem.freq = int(_data.rx_freq)*10
                mem.offset = int(_data.tx_freq)*10
                if isvfo:
                    # we want split, so we have to reconstruct it
                    # from -/0/+ modes
                    if mem.duplex == '-':
                        mem.offset = mem.freq - mem.offset
                    elif mem.duplex == '':
                        mem.offset = mem.freq
                    elif mem.duplex == '+':
                        mem.offset = mem.freq + mem.offset
                    mem.duplex = 'split'
                else:
                    # we want -/0/+, so we have to reconstruct it
                    # from split modes
                    if mem.freq > mem.offset:
                        mem.offset = mem.freq - mem.offset
                        mem.duplex = '-'
                    elif mem.freq < mem.offset:
                        mem.offset = mem.offset - mem.freq
                        mem.duplex = '+'
                    else:
                        mem.offset = 0
                        mem.duplex = ''
######################################################################

        return mem

    def set_memory(self, mem):
        (index, designator,
         _data, _name, _present, _priority,
         isregular, isvfo, iscall) = self._get_memory_structs(mem.number)

        # handle empty channels
        if mem.empty:
            if isregular:
                _present.set_value(False)
                _priority.set_value(False)
                self._uninitialize(_data, 16)
                self._uninitialize(_name, 6)
            else:
                raise errors.InvalidValueError(
                        "Can't remove CALL and/or VFO channels!")
            return

        # handle regular channel stuff like name and present+priority flags
        if isregular:
            if not bool(_present):
                self._set_sane_defaults(_data)
            _name.set_value(
                    encode_ffstring(self.filter_name(mem.name.rstrip()),
                                    len(_name)))
            _present.set_value(True)
            _priority.set_value(1-SKIP_MODES.index(mem.skip))

        # frequency data
        rxf = int(mem.freq/10)
        txf = int(mem.offset/10)

        _data.rx_freq.set_value(rxf)

        if isvfo:
            # fake split modes on write, for channels
            # that do not support it, which are some
            # (the two vfo channels)
            if mem.duplex == 'split':
                for band in BANDS:
                    rb = mem.freq in range(band[0], band[1])
                    tb = mem.offset in range(band[0], band[1])
                    if rb != tb:
                        raise errors.InvalidValueError(
                                "VFO frequencies should be on the same band")
                if rxf < txf:
                    _data.duplex_sign.set_value(1)
                    _data.tx_freq.set_value(txf - rxf)
                elif rxf > txf:
                    _data.duplex_sign.set_value(2)
                    _data.tx_freq.set_value(rxf-txf)
                else:
                    _data.duplex_sign.set_value(0)
                    _data.tx_freq.set_value(0)
            else:
                _data.duplex_sign.set_value(DUPLEX_MODES.index(mem.duplex))
                _data.tx_freq.set_value(txf)

######################################################################
            if ENABLE_DANGEROUS_EXPERIMENTAL_FEATURES:
                if mem.duplex == 'split':
                    _data.experimental_duplex_mode_indicator.set_value(0)
                else:
                    _data.experimental_duplex_mode_indicator.set_value(1)
######################################################################

        else:
            # fake duplex modes on write, for channels
            # that do not support it, which are most
            # (all the memory channels)
            if mem.duplex == '' or mem.duplex is None:
                _data.tx_freq.set_value(rxf)
            elif mem.duplex == '+':
                _data.tx_freq.set_value(rxf + txf)
            elif mem.duplex == '-':
                _data.tx_freq.set_value(rxf - txf)
            else:
                _data.tx_freq.set_value(txf)

######################################################################
            if ENABLE_DANGEROUS_EXPERIMENTAL_FEATURES:
                if mem.duplex != 'split':
                    _data.experimental_duplex_mode_indicator.set_value(0)
                else:
                    _data.experimental_duplex_mode_indicator.set_value(1)
######################################################################

        # tone data
        tonedata = chirp_common.split_tone_encode(mem)
        for i in range(2):
            dihl = unparse_tone(tonedata[i])
            if dihl is not None:
                _data.tone[i].digital.set_value(dihl[0])
                _data.tone[i].invert.set_value(dihl[1])
                _data.tone[i].high.set_value(dihl[2])
                _data.tone[i].low.set_value(dihl[3])
            else:
                _data.tone[i].digital.set_value(1)
                _data.tone[i].invert.set_value(1)
                _data.tone[i].high.set_value(0x3f)
                _data.tone[i].low.set_value(0xff)

######################################################################
        if ENABLE_DANGEROUS_EXPERIMENTAL_FEATURES:
            if mem.tmode == 'Cross' and mem.cross_mode in ['Tone->',
                                                           'Tone->Tone',
                                                           'DTCS->DTCS']:
                _data.experimental_cross_mode_indicator.set_value(0)
            else:
                _data.experimental_cross_mode_indicator.set_value(1)
######################################################################

        # transmit mode and power level
        _data.modulation_width.set_value(MODES.index(mem.mode))
        if str(mem.power) == 'High':
            _data.txpower.set_value(1)
        else:
            _data.txpower = 0

    def get_settings(self):
        _model = self._memobj.mem.model_information
        _settings = self._memobj.mem.opt_settings
        _ptt_id_data = self._memobj.mem.ptt_id_data
        _msk_settings = self._memobj.mem.msk_settings
        _dtmf_settings = self._memobj.mem.dtmf_settings
        _5tone_settings = self._memobj.mem.five_tone_settings
        _broadcast = self._memobj.mem.fm_radio

        # for safety reasons we are showing these as read-only
        model_unit_settings = [
            integer_setting("vhflo", "VHF lower bound",
                            _model.band_limits[0].lower_freq,
                            134, 176,
                            int_from_mem=lambda x:int(int(x)/10),
                            mem_from_int=None),
            integer_setting("vhfhi", "VHF upper bound",
                            _model.band_limits[0].upper_freq,
                            134, 176,
                            int_from_mem=lambda x:int(int(x)/10),
                            mem_from_int=None),
            integer_setting("uhflo", "UHF lower bound",
                            _model.band_limits[1].lower_freq,
                            400, 480,
                            int_from_mem=lambda x:int(int(x)/10),
                            mem_from_int=None),
            integer_setting("uhfhi", "UHF upper bound",
                            _model.band_limits[1].upper_freq,
                            400, 480,
                            int_from_mem=lambda x:int(int(x)/10),
                            mem_from_int=None),
            ff_string_setting("model", "Model string",
                              _model.model_string,
                              0, 6)

        ]
        for s in model_unit_settings:
            s.value.set_mutable(False)
        model_unit_settings.append(ff_string_setting(
            "info", "Unit Information",
            self._memobj.mem.radio_information_string,
            0, 16))

        # tx/rx related settings
        radio_channel_settings = [
            list_setting("vfostep", "VFO step size",
                         _settings.vfo_step,
                         VFO_STRIDE),
            list_setting("abwatch", "Main watch",
                         _settings.main_watch,
                         AB),
            list_setting("watchmade", "Watch mode",
                         _settings.main_watch,
                         WATCH_MODES),
            list_setting("amode", "A mode",
                         _settings.workmode_a,
                         AB_MODES),
            list_setting("bmode", "B mode",
                         _settings.workmode_b,
                         AB_MODES),
            integer_setting("achan", "A channel index",
                            _settings.channel_a,
                            1, 128,
                            int_from_mem=lambda i:i+1,
                            mem_from_int=lambda i:i-1),
            integer_setting("bchan", "B channel index",
                            _settings.channel_b,
                            1, 128,
                            int_from_mem=lambda i:i+1,
                            mem_from_int=lambda i:i-1),
            integer_setting("pchan", "Priority channel index",
                            _settings.priority_channel,
                            1, 128, int_from_mem=lambda i:i+1,
                            mem_from_int=lambda i:i-1),
            list_setting("cactive", "Call channel active?",
                         _settings.call_channel_active,
                         NO_YES),
            list_setting("scanm", "Scan mode",
                         _settings.scan_mode,
                         SCAN_MODES),
            list_setting("swait", "Wait time",
                         _settings.wait_time,
                         WAIT_TIMES),
            # it is unclear what this option below does,
            # possibly squelch tail elimination?
            list_setting("tail", "Relay without disable tail (?)",
                         _settings.relay_without_disable_tail,
                         NO_YES),
            list_setting("batsav", "Battery saving mode",
                         _settings.battery_save,
                         OFF_ON),
            ]

        # user interface related settings
        interface_settings = [
            list_setting("sidehold", "Side button hold action",
                         _settings.side_button_hold_mode,
                         BUTTON_MODES),
            list_setting("sideclick", "Side button click action",
                         _settings.side_button_click_mode,
                         BUTTON_MODES),
            list_setting("bootmt", "Boot message type",
                         _settings.boot_message_mode,
                         BOOT_MESSAGE_TYPES),
            ff_string_setting("bootm", "Boot message",
                              _settings.boot_message,
                              0, 6),
            list_setting("beep", "Key beep",
                         _settings.key_beep,
                         OFF_ON),
            list_setting("talkback", "Menu talkback",
                         _settings.voice_announce,
                         TALKBACK),
            list_setting("sidetone", "DTMF sidetone",
                         _settings.dtmf_sidetone,
                         OFF_ON),
            list_setting("roger", "Roger beep",
                         _settings.use_roger_beep,
                         ROGER_BEEP),
            list_setting("backlm", "Backlight mode",
                         _settings.backlight_mode,
                         BACKLIGHT_MODES),
            list_setting("backlc", "Backlight color",
                         _settings.backlight_color,
                         BACKLIGHT_COLORS),
            integer_setting("squelch", "Squelch level",
                            _settings.squelch_level,
                            0, 9),
            list_setting("voxg", "Vox gain",
                         _settings.vox_gain,
                         VOX_GAIN),
            list_setting("voxd", "Vox delay",
                         _settings.vox_delay,
                         VOX_DELAYS),
            list_setting("txal", "Trinsmit time alarm",
                         _settings.tx_timeout,
                         TRANSMIT_ALARMS),
            ]

        # settings related to tone/data sending and interpretation
        data_general_settings = [
            list_setting("disptt", "Display PTT ID",
                         _settings.dis_ptt_id,
                         NO_YES),
            list_setting("pttidt", "PTT ID signal type",
                         _settings.ptt_id_type,
                         DATA_MODES)
            ]

        data_msk_settings = [
            ff_string_setting("bot", "MSK PTT ID (BOT)",
                              _ptt_id_data[0].entry,
                              0, 6, autowrite=False),
            ff_string_setting("eot", "MSK PTT ID (EOT)",
                              _ptt_id_data[1].entry,
                              0, 6, autowrite=False),
            ff_string_setting("id", "MSK ID code",
                              _msk_settings.id_code,
                              0, 4, charset=HEXADECIMAL),
            list_setting("mskr", "MSK reverse",
                         _settings.msk_reverse,
                         NO_YES)
            ]

        data_dtmf_settings = [
            dtmf_string_setting("bot", "DTMF PTT ID (BOT)",
                                _ptt_id_data[0].entry,
                                _ptt_id_data[0].length,
                                0, 8, autowrite=False),
            dtmf_string_setting("eot", "DTMF PTT ID (EOT)",
                                _ptt_id_data[1].entry,
                                _ptt_id_data[1].length,
                                0, 8, autowrite=False),
            dtmf_string_setting("id",  "DTMF ID code",
                                _dtmf_settings.id_code,
                                _dtmf_settings.id_code_length,
                                3, 8),

            integer_setting("time",   "Digit time (ms)",
                            _dtmf_settings.timing.digit_length,
                            50,  200, step=10,
                            int_from_mem=lambda x:x*10,
                            mem_from_int=lambda x:int(x/10)),
            integer_setting("pause",  "Inter digit time (ms)",
                            _dtmf_settings.timing.digit_length,
                            50,  200, step=10,
                            int_from_mem=lambda x:x*10,
                            mem_from_int=lambda x:int(x/10)),
            integer_setting("time1",  "First digit time (ms)",
                            _dtmf_settings.timing.digit_length,
                            50,  200, step=10,
                            int_from_mem=lambda x:x*10,
                            mem_from_int=lambda x:int(x/10)),
            integer_setting("pause1", "First digit delay (ms)",
                            _dtmf_settings.timing.digit_length,
                            100, 1000, step=50,
                            int_from_mem=lambda x:x*50,
                            mem_from_int=lambda x:int(x/50)),

            list_setting("arst", "Auto reset time",
                         _dtmf_settings.reset_time,
                         DTMF_TONE_RESET_TIME),
            list_setting("grp", "Group code",
                         _dtmf_settings.group_code,
                         DTMF_GROUPS),
            dtmf_string_setting("stunt", "TX Stun code",
                                _dtmf_settings.tx_stun_code,
                                _dtmf_settings.tx_stun_code_length,
                                3, 8),
            dtmf_string_setting("cstunt", "TX Stun cancel code",
                                _dtmf_settings.cancel_tx_stun_code,
                                _dtmf_settings.cancel_tx_stun_code_length,
                                3, 8),
            dtmf_string_setting("stunrt", "RX/TX Stun code",
                                _dtmf_settings.rxtx_stun_code,
                                _dtmf_settings.rxtx_stun_code_length,
                                3, 8),
            dtmf_string_setting("cstunrt", "RX/TX Stun cancel code",
                                _dtmf_settings.cancel_rxtx_stun_code,
                                _dtmf_settings.cancel_rxtx_stun_code_length,
                                3, 8),
            list_setting("altr", "Alert/Transpond",
                         _dtmf_settings.alert_transpond,
                         DTMF_ALERT_TRANSPOND),
            ]

        data_5tone_settings = [
            five_tone_string_setting("bot",
                                     "5-Tone PTT ID (BOT)",
                                     _ptt_id_data[0].entry,
                                     autowrite=False),
            five_tone_string_setting("eot",
                                     "5-Tone PTT ID (EOT)",
                                     _ptt_id_data[1].entry,
                                     autowrite=False),
            five_tone_string_setting("id",
                                     "5-tone ID code",
                                     _5tone_settings.id_code),
            list_setting("arst", "Auto reset time",
                         _5tone_settings.reset_time,
                         TONE_RESET_TIME),
            five_tone_string_setting("stunt",
                                     "TX Stun code",
                                     _5tone_settings.tx_stun_code),
            five_tone_string_setting("cstunt",
                                     "TX Stun cancel code",
                                     _5tone_settings.cancel_tx_stun_code),
            five_tone_string_setting("stunrt",
                                     "RX/TX Stun code",
                                     _5tone_settings.rxtx_stun_code),
            five_tone_string_setting("cstunrt",
                                     "RX/TX Stun cancel code",
                                     _5tone_settings.cancel_rxtx_stun_code),
            list_setting("altr", "Alert/Transpond",
                         _5tone_settings.alert_transpond,
                         FIVE_TONE_ALERT_TRANSPOND),
            list_setting("std", "5-Tone standard",
                         _5tone_settings.tone_standard,
                         FIVE_TONE_STANDARDS),
            ]
        for i in range(4):
            s = ['z1', 'z2', 'c1', 'ct'][i]
            l = FIVE_TONE_STANDARDS[i]
            data_5tone_settings.append(
                settings.RadioSettingGroup(
                    s, '%s settings' % l,
                    integer_setting("%speriod" % s, "%s Period (ms)" % l,
                                    _5tone_settings.tone_settings[i].period,
                                    20, 255),
                    list_setting("%sgrp" % s, "%s Group code" % l,
                                 _5tone_settings.tone_settings[i].group_code,
                                 HEXADECIMAL),
                    list_setting("%srpt" % s, "%s Repeat code" % l,
                                 _5tone_settings.tone_settings[i].repeat_code,
                                 HEXADECIMAL)))

        data_msk_call_list = []
        data_dtmf_call_list = []
        data_5tone_call_list = []
        for i in range(9):
            j = i+1
            data_msk_call_list.append(
                ff_string_setting("ce%d" % i,
                                  "MSK call entry %d" % j,
                                  _msk_settings.phone_book[i].entry,
                                  0, 4,
                                  charset=HEXADECIMAL))

            data_dtmf_call_list.append(
                dtmf_string_setting("ce%d" % i,
                                    "DTMF call entry %d" % j,
                                    _dtmf_settings.phone_book[i].entry,
                                    _dtmf_settings.phone_book[i].length,
                                    0, 10))

            data_5tone_call_list.append(
                five_tone_string_setting("ce%d" % i,
                                         "5-Tone call entry %d" % j,
                                         _5tone_settings.phone_book[i].entry)),

        data_settings = data_general_settings
        data_settings.extend([
                settings.RadioSettingGroup("MSK_s", "MSK settings",
                                           *data_msk_settings),
                settings.RadioSettingGroup("MSK_c", "MSK call list",
                                           *data_msk_call_list),
                settings.RadioSettingGroup("DTMF_s", "DTMF settings",
                                           *data_dtmf_settings),
                settings.RadioSettingGroup("DTMF_c", "DTMF call list",
                                           *data_dtmf_call_list),
                settings.RadioSettingGroup("5-Tone_s", "5-tone settings",
                                           *data_5tone_settings),
                settings.RadioSettingGroup("5-Tone_c", "5-tone call list",
                                           *data_5tone_call_list)
            ])

        # settings related to the various ways the radio can be locked down
        locking_settings = [
            list_setting("autolock", "Automatic timed keypad lock",
                         _settings.auto_keylock,
                         OFF_ON),
            list_setting("lockon", "Current status of keypad lock",
                         _settings.keypad_lock,
                         INACTIVE_ACTIVE),
            list_setting("nokeypad", "Disable keypad",
                         _settings.allow_keypad,
                         YES_NO),
            list_setting("rxstun", "Disable receiver (rx stun)",
                         _settings.rx_stun,
                         NO_YES),
            list_setting("txstun", "Disable transmitter (tx stun)",
                         _settings.tx_stun,
                         NO_YES),
            ]

        # broadcast FM radio settings
        broadcast_settings = [
            list_setting("band", "Frequency interval",
                         _broadcast.receive_range,
                         BFM_BANDS),
            list_setting("stride", "VFO step",
                         _broadcast.channel_stepping,
                         BFM_STRIDE),
            frequency_setting("vfo", "VFO frequency (MHz)",
                              _broadcast.vfo_freq)
        ]

        for i in range(10):
            broadcast_settings.append(
                frequency_setting("bcd%d" % i, "Memory %d frequency" % i,
                                  _broadcast.memory[i].entry))

        return settings.RadioSettings(
            settings.RadioSettingGroup("model", "Model/Unit information",
                                       *model_unit_settings),
            settings.RadioSettingGroup("radio", "Radio/Channel settings",
                                       *radio_channel_settings),
            settings.RadioSettingGroup("interface", "Interface",
                                       *interface_settings),
            settings.RadioSettingGroup("data", "Data",
                                       *data_settings),
            settings.RadioSettingGroup("locking", "Locking",
                                       *locking_settings),
            settings.RadioSettingGroup("broadcast",
                                       "Broadcast FM radio settings",
                                       *broadcast_settings)
        )

    def set_settings(self, s, parent=''):
        # The helper classes take care of all settings except these below,
        # since it is a single instance of settings having interdependencies,
        # i.e., which value that gets written to the memory depends on
        # the value of another setting. The value of the ptt id type setting
        # decides which of the msk/dtmf/5-tone ptt id strings are
        # actually written to memory.
        ds = sbyn(s, 'data')
        idts = sbyn(ds, 'pttidt').value
        idtv = idts.get_value()
        cs = sbyn(ds, idtv+'_s')
        tss = [sbyn(cs, e).value for e in ['bot', 'eot']]

        for ts in tss:
            ts.write_mem()



# Copyright 2016-2023:
# * Jim Unroe KC9HI, <rock.unroe@gmail.com>
# Modified for Baojie BJ-218: 2018
#    Rick DeWitt (RJD), <aa0rd@yahoo.com>
# Modified for Baojie BJ-318: April 2021
#    Mark Hartong, AJ4YI <mark.hartong@verizon.net>
# Modified for Baojie BJ-318: January 2023
#    Mark Hartong, AJ4YI <mark.hartong@verizon.net>
#    1. Removed Experimental Warning
#    2. Added Brown & Yellow as Color Selections
#    3. Max VFO Volume Setting to 15 ( vice 10)
# Modified UHF band limits: February 2023
#    Jim Unroe, KC9HI <rock.unroe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import struct
import logging

from chirp import chirp_common, directory, memmap
from chirp import bitwise, errors, util
from chirp.settings import RadioSettingGroup, RadioSetting, \
    RadioSettingValueBoolean, RadioSettingValueList, \
    RadioSettingValueString, RadioSettingValueInteger, \
    RadioSettingValueFloat, RadioSettings, InvalidValueError

LOG = logging.getLogger(__name__)

MEM_FORMAT = """
#seekto 0x0200;
struct {
  u8  init_bank; // determines which VFO is primary A or B
  u8  volume;    // not used BJ-318, band volume is controlled vfo u8 bvol
  u16 fm_freq;   // not used BJ-318, freq is controlled hello_lims u16 fm_318
  u8  wtled;     // not used BJ-318
  u8  rxled;     // not used BJ-318
  u8  txled;     // not used BJ-318
  u8  ledsw;
  u8  beep;
  u8  ring;
  u8  bcl;
  u8  tot;
  u16 sig_freq;
  u16 dtmf_txms;
  u8  init_sql;  // not used BJ-318, band squelch is controlled vfo u8 sql
  u8  rptr_mode;
} settings;

#seekto 0x0240;
struct {
  u8  dtmf1_cnt;
  u8  dtmf1[7];
  u8  dtmf2_cnt;
  u8  dtmf2[7];
  u8  dtmf3_cnt;
  u8  dtmf3[7];
  u8  dtmf4_cnt;
  u8  dtmf4[7];
  u8  dtmf5_cnt;
  u8  dtmf5[7];
  u8  dtmf6_cnt;
  u8  dtmf6[7];
  u8  dtmf7_cnt;
  u8  dtmf7[7];
  u8  dtmf8_cnt;
  u8  dtmf8[7];
} dtmf_tab;

#seekto 0x0280;
struct {
  u8  native_id_cnt;
  u8  native_id_code[7];
  u8  master_id_cnt;
  u8  master_id_code[7];
  u8  alarm_cnt;
  u8  alarm_code[5];
  u8  id_disp_cnt;
  u8  id_disp_code[5];
  u8  revive_cnt;
  u8  revive_code[5];
  u8  stun_cnt;
  u8  stun_code[5];
  u8  kill_cnt;
  u8  kill_code[5];
  u8  monitor_cnt;
  u8  monitor_code[5];
  u8  state_now;
} codes;

#seekto 0x02d0;
struct {
  u8  hello1_cnt;    // not used in BJ-318, is set in memory map
  char  hello1[7];   // not used in BJ-318, is set in memory map
  u8  hello2_cnt;    // not used in BJ-318, is set in memory map
  char  hello2[7];   // not used in BJ-318, is set in memory map
  u32  vhf_low;
  u32  vhf_high;
  u32  uhf_low;
  u32  uhf_high;
  u8  lims_on;       // first byte @ at address 0x02f0;
  u8 unknown_1;      // use in BJ-318 unknown
  u8 lang;           // BJ-318 Display language
                     // 02 = English 00 and 01 are Asian
  u8 up_scr_color;   // BJ-318 upper screen character color
  u8 dn_scr_color;   // BJ-318 lower screen character color
                     // purple=00, red=01, emerald=02, blue=03,
                     // sky_blue=04, black=05, brown=06, yellow=07 required
                     // hex values for color
                     // Note   sky_blue and black look the same on the
                     // BJ-318 screen
  u16 fm_318;        // FM stored Frequency in BJ-318
} hello_lims;

struct vfo {
  u8  frq_chn_mode;
  u8  chan_num;
  u32 rxfreq;
  u16 is_rxdigtone:1,
      rxdtcs_pol:1,
      rx_tone:14;
  u8  rx_mode;
  u8  unknown_ff;
  u16 is_txdigtone:1,
      txdtcs_pol:1,
      tx_tone:14;
  u8  launch_sig;
  u8  tx_end_sig;
  u8  bpower;        // sets power Hi=02h, Medium=01h, Low=00
                     // sets power for entire vfo band
  u8  fm_bw;
  u8  cmp_nder;
  u8  scrm_blr;
  u8  shift;
  u32 offset;
  u16 step;
  u8  sql;           // squelch for entire vfo band
                     // integer values 0 (low) to 9 (high)
  u8  bvol;          // sets volume for vfo band
                     // integer values 0  (low) TO 15 (high)
};

#seekto 0x0300;
struct {
  struct vfo vfoa;
} upper;

#seekto 0x0380;
struct {
  struct vfo vfob;
} lower;

struct mem {
  u32 rxfreq;
  u16 is_rxdigtone:1,
      rxdtcs_pol:1,
      rxtone:14;
  u8  recvmode;
  u32 txfreq;
  u16 is_txdigtone:1,
      txdtcs_pol:1,
      txtone:14;
  u8  botsignal;
  u8  eotsignal;
  u8  power:1,            // binary value for
                          // individual channel power
                          // set to "High" or "Low"
                          // BJ-318 band power overrides any
                          // individual channel power setting
      wide:1,
      compander:1,
      scrambler:1,
      unknown:4;
  u8  namelen;
  u8  name[7];
};

#seekto 0x0400;
struct mem upper_memory[128];

#seekto 0x1000;
struct mem lower_memory[128];

#seekto 0x1C00;
struct {
  char  mod_num[6];
} mod_id;
"""

MEM_SIZE = 0x1C00
BLOCK_SIZE = 0x40
STIMEOUT = 2
LIST_RECVMODE = ["QT/DQT", "QT/DQT + Signaling"]
LIST_SIGNAL = ["Off"] + ["DTMF%s" % x for x in range(1, 9)] + \
              ["DTMF%s + Identity" % x for x in range(1, 9)] + \
              ["Identity code"]
# Band Power settings, can be different than channel power
LIST_BPOWER = ["Low", "Mid", "High"]    # Tri-power models

# Screen Color Settings
# BJ-218 Colors
LIST_COLOR = ["Off", "Orange", "Blue", "Purple"]
# BJ-318 Colors
LIST_COLOR318 = ["Purple", "Red", "Emerald", "Blue", "Sky_Blue", "Black",
                 "Brown", "Yellow"]
LIST_LEDSW = ["Auto", "On"]
LIST_RING = ["Off"] + ["%s" % x for x in range(1, 10)]
LIST_TDR_DEF = ["A-Upper", "B-Lower"]
LIST_TIMEOUT = ["Off"] + ["%s" % x for x in range(30, 630, 30)]
LIST_VFOMODE = ["Frequency Mode", "Channel Mode"]
# Tones are numeric, Defined in \chirp\chirp_common.py
TONES_CTCSS = sorted(chirp_common.TONES)
# Converted to strings
LIST_CTCSS = ["Off"] + [str(x) for x in TONES_CTCSS]
# Now append the DxxxN and DxxxI DTCS codes from chirp_common
for x in chirp_common.DTCS_CODES:
    LIST_CTCSS.append("D{:03d}N".format(x))
for x in chirp_common.DTCS_CODES:
    LIST_CTCSS.append("D{:03d}R".format(x))
LIST_BW = ["Narrow", "Wide"]
LIST_SHIFT = ["Off", " + ", " - "]
STEPS = [2.5, 5.0, 6.25, 10.0, 12.5, 20.0, 25.0, 50.0]
LIST_STEPS = [str(x) for x in STEPS]
LIST_STATE = ["Normal", "Stun", "Kill"]
LIST_SSF = ["1000", "1450", "1750", "2100"]
LIST_DTMFTX = ["50", "100", "150", "200", "300", "500"]


def _clean_buffer(radio):
    radio.pipe.timeout = 0.005
    junk = radio.pipe.read(256)
    radio.pipe.timeout = STIMEOUT
    if junk:
        LOG.debug("Got %i bytes of junk before starting" % len(junk))


def _rawrecv(radio, amount):
    """Raw read from the radio device"""
    data = b""
    try:
        data = radio.pipe.read(amount)
    except:
        _exit_program_mode(radio)
        msg = "Generic error reading data from radio; check your cable."
        raise errors.RadioError(msg)

    if len(data) != amount:
        _exit_program_mode(radio)
        msg = "Error reading from radio: not the amount of data we want."
        raise errors.RadioError(msg)

    return data


def _rawsend(radio, data):
    """Raw send to the radio device"""
    try:
        radio.pipe.write(data)
    except:
        raise errors.RadioError("Error sending data to radio")


def _make_frame(cmd, addr, length, data=""):
    """Pack the info in the header format"""
    frame = struct.pack(">4sHH", cmd, addr, length)
    # Add the data if set
    if len(data) != 0:
        frame += data
    # Return the data
    return frame


def _recv(radio, addr, length):
    """Get data from the radio """

    data = _rawrecv(radio, length)

    # DEBUG
    LOG.info("Response:")
    LOG.debug(util.hexprint(data))

    return data


def _do_ident(radio):
    """Put the radio in PROGRAM mode & identify it"""
    # Set the serial discipline
    radio.pipe.baudrate = 19200
    radio.pipe.parity = "N"
    radio.pipe.timeout = STIMEOUT

    # Flush input buffer
    _clean_buffer(radio)

    magic = b"PROM_LIN"

    _rawsend(radio, magic)

    ack = _rawrecv(radio, 1)
    if ack != b"\x06":
        _exit_program_mode(radio)
        if ack:
            LOG.debug(repr(ack))
        raise errors.RadioError("Radio did not respond")

    return True


def _exit_program_mode(radio):
    endframe = b"EXIT"
    _rawsend(radio, endframe)


def _download(radio):
    """Get the memory map"""

    # Put radio in program mode and identify it
    _do_ident(radio)

    # UI progress
    status = chirp_common.Status()
    status.cur = 0
    status.max = MEM_SIZE // BLOCK_SIZE
    status.msg = "Cloning from radio..."
    radio.status_fn(status)

    data = b""
    for addr in range(0, MEM_SIZE, BLOCK_SIZE):
        frame = _make_frame(b"READ", addr, BLOCK_SIZE)
        # DEBUG
        LOG.info("Request sent:")
        LOG.debug(util.hexprint(frame))

        # Sending the read request
        _rawsend(radio, frame)

        # Now we read
        d = _recv(radio, addr, BLOCK_SIZE)

        # Aggregate the data
        data += d

        # UI Update
        status.cur = addr // BLOCK_SIZE
        status.msg = "Cloning from radio..."
        radio.status_fn(status)

    _exit_program_mode(radio)

    return data


def _upload(radio):
    """Upload procedure"""

    # Put radio in program mode and identify it
    _do_ident(radio)

    # UI progress
    status = chirp_common.Status()
    status.cur = 0
    status.max = MEM_SIZE // BLOCK_SIZE
    status.msg = "Cloning to radio..."
    radio.status_fn(status)

    # The fun starts here
    for addr in range(0, MEM_SIZE, BLOCK_SIZE):
        # Sending the data
        data = radio.get_mmap()[addr:addr + BLOCK_SIZE]

        frame = _make_frame(b"WRIE", addr, BLOCK_SIZE, data)

        _rawsend(radio, frame)

        # Receiving the response
        ack = _rawrecv(radio, 1)
        if ack != b"\x06":
            _exit_program_mode(radio)
            msg = "Bad ack writing block 0x%04x" % addr
            raise errors.RadioError(msg)

        # UI Update
        status.cur = addr // BLOCK_SIZE
        status.msg = "Cloning to radio..."
        radio.status_fn(status)

    _exit_program_mode(radio)


def model_match(cls, data):
    """Match the opened/downloaded image to the correct version"""
    if len(data) == 0x1C08:
        rid = data[0x1C00:0x1C08]
        return rid.startswith(cls.MODEL.encode())
    else:
        return False


def _split(rf, f1, f2):
    """Returns False if the two freqs are in the same band (no split)
    or True otherwise"""

    # Determine if the two freqs are in the same band
    for low, high in rf.valid_bands:
        if f1 >= low and f1 <= high and \
                f2 >= low and f2 <= high:
            # If the two freqs are on the same Band this is not a split
            return False

    # If you get here is because the freq pairs are split
    return True


@directory.register
class LT725UV(chirp_common.CloneModeRadio):
    """LUITON LT-725UV Radio"""
    VENDOR = "LUITON"
    MODEL = "LT-725UV"
    NEEDS_COMPAT_SERIAL = False
    MODES = ["NFM", "FM"]
    TONES = chirp_common.TONES
    DTCS_CODES = tuple(sorted(chirp_common.DTCS_CODES + (645,)))
    NAME_LENGTH = 7
    DTMF_CHARS = list("0123456789ABCD*#")

    # Channel Power: 2 levels
    # BJ-318 uses 3 power levels for each VFO "Band"
    # Low = 5W, Med = 10W, High = 25W
    # The band power selection in a VFO applies to the VFO and overrides
    # the stored channel power selection
    # The firmware channel memory structure provides only 1 bit for
    # individual channel power settings, limiting potential channel
    # power selection options to 2 levels: Low or High.
    POWER_LEVELS = [chirp_common.PowerLevel("Low", watts=5.00),
                    chirp_common.PowerLevel("High", watts=30.00)]

    VALID_BANDS = [(136000000, 176000000),
                   (400000000, 490000000)]

    # Valid chars on the LCD
    VALID_CHARS = chirp_common.CHARSET_ALPHANUMERIC + \
        "`{|}!\"#$%&'()*+,-./:;<=>?@[]^_"

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        if cls.MODEL == "BJ-318":
            msg = \
                ('\n'
                 'The individual BJ-318 channel power settings set in CHIRP'
                 ' are ignored by \n'
                 'the radio. They are allowed to be set in hopes of a future'
                 ' firmware \n'
                 'change by the manufacturer. While the CHIRP driver will'
                 ' allow setting \n'
                 'the individual VFO power to High (25W), Med (10W) or Low'
                 ' (5W) they \n'
                 'are converted to just Low (5W) and High (25W) on upload. The'
                 ' Med setting \n'
                 'reverts to Low. \n'
                 '\n'
                 'Changing a channel power setting must be done manually using'
                 ' the front \n'
                 'panel menu control (or the microphone controls while in VFO'
                 ' mode.)'
                 )

        else:
            msg = \
                ('Some notes about POWER settings:\n'
                 '- The individual channel power settings are ignored'
                 ' by the radio.\n'
                 '  They are allowed to be set (and downloaded) in hopes of'
                 ' a future firmware update.\n'
                 '- Power settings done \'Live\' in the radio apply to the'
                 ' entire upper or lower band.\n'
                 '- Tri-power radio models will set and download the three'
                 ' band-power'
                 ' levels, but they are converted to just Low and High at'
                 ' upload.'
                 ' The Mid setting reverts to Low.'
                 )

        rp.info = msg

        rp.pre_download = _(
            "Follow this instructions to download your info:\n"
            "1 - Turn off your radio\n"
            "2 - Connect your interface cable\n"
            "3 - Turn on your radio\n"
            "4 - Do the download of your radio data\n")
        rp.pre_upload = _(
            "Follow this instructions to upload your info:\n"
            "1 - Turn off your radio\n"
            "2 - Connect your interface cable\n"
            "3 - Turn on your radio\n"
            "4 - Do the upload of your radio data\n")
        return rp

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_bank = False
        rf.has_tuning_step = False
        rf.can_odd_split = True
        rf.has_name = True
        rf.has_offset = True
        rf.has_mode = True
        rf.has_dtcs = True
        rf.has_rx_dtcs = True
        rf.has_dtcs_polarity = True
        rf.has_ctone = True
        rf.has_cross = True
        rf.has_sub_devices = self.VARIANT == ""
        rf.valid_modes = self.MODES
        rf.valid_characters = self.VALID_CHARS
        rf.valid_duplexes = ["", "-", "+", "split", "off"]
        rf.valid_tmodes = ['', 'Tone', 'TSQL', 'DTCS', 'Cross']
        rf.valid_cross_modes = [
            "Tone->Tone",
            "DTCS->",
            "->DTCS",
            "Tone->DTCS",
            "DTCS->Tone",
            "->Tone",
            "DTCS->DTCS"]
        rf.valid_skips = []
        rf.valid_power_levels = self.POWER_LEVELS
        rf.valid_name_length = self.NAME_LENGTH
        rf.valid_dtcs_codes = self.DTCS_CODES
        rf.valid_bands = self.VALID_BANDS
        rf.memory_bounds = (1, 128)
        rf.valid_tuning_steps = STEPS
        return rf

    def get_sub_devices(self):
        return [LT725UVUpper(self._mmap), LT725UVLower(self._mmap)]

    def sync_in(self):
        """Download from radio"""
        try:
            data = _download(self)
        except errors.RadioError:
            # Pass through any real errors we raise
            raise
        except:
            # If anything unexpected happens, make sure we raise
            # a RadioError and log the problem
            LOG.exception('Unexpected error during download')
            raise errors.RadioError('Unexpected error communicating '
                                    'with the radio')
        self._mmap = memmap.MemoryMapBytes(data)
        self.process_mmap()

    def sync_out(self):
        """Upload to radio"""
        try:
            _upload(self)
        except:
            # If anything unexpected happens, make sure we raise
            # a RadioError and log the problem
            LOG.exception('Unexpected error during upload')
            raise errors.RadioError('Unexpected error communicating '
                                    'with the radio')

    def process_mmap(self):
        """Process the mem map into the mem object"""
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number - 1])

    def _memory_obj(self, suffix=""):
        return getattr(self._memobj, "%s_memory%s" % (self._vfo, suffix))

    def _get_dcs(self, val):
        return int(str(val)[2:-18])

    def _set_dcs(self, val):
        return int(str(val), 16)

    def get_memory(self, number):
        _mem = self._memory_obj()[number - 1]

        mem = chirp_common.Memory()
        mem.number = number

        if _mem.get_raw(asbytes=False)[0] == "\xff":
            mem.empty = True
            return mem

        mem.freq = int(_mem.rxfreq) * 10

        if _mem.txfreq == 0xFFFFFFFF:
            # TX freq not set
            mem.duplex = "off"
            mem.offset = 0
        elif int(_mem.rxfreq) == int(_mem.txfreq):
            mem.duplex = ""
            mem.offset = 0
        elif _split(self.get_features(), mem.freq, int(_mem.txfreq) * 10):
            mem.duplex = "split"
            mem.offset = int(_mem.txfreq) * 10
        else:
            mem.duplex = int(_mem.rxfreq) > int(_mem.txfreq) and "-" or "+"
            mem.offset = abs(int(_mem.rxfreq) - int(_mem.txfreq)) * 10

        for char in _mem.name[0:_mem.namelen]:
            mem.name += chr(char)

        dtcs_pol = ["N", "N"]

        if _mem.rxtone == 0x3FFF:
            rxmode = ""
        elif _mem.is_rxdigtone == 0:
            # CTCSS
            rxmode = "Tone"
            mem.ctone = int(_mem.rxtone) / 10.0
        else:
            # Digital
            rxmode = "DTCS"
            mem.rx_dtcs = self._get_dcs(_mem.rxtone)
            if _mem.rxdtcs_pol == 1:
                dtcs_pol[1] = "R"

        if _mem.txtone == 0x3FFF:
            txmode = ""
        elif _mem.is_txdigtone == 0:
            # CTCSS
            txmode = "Tone"
            mem.rtone = int(_mem.txtone) / 10.0
        else:
            # Digital
            txmode = "DTCS"
            mem.dtcs = self._get_dcs(_mem.txtone)
            if _mem.txdtcs_pol == 1:
                dtcs_pol[0] = "R"

        if txmode == "Tone" and not rxmode:
            mem.tmode = "Tone"
        elif txmode == rxmode and txmode == "Tone" and mem.rtone == mem.ctone:
            mem.tmode = "TSQL"
        elif txmode == rxmode and txmode == "DTCS" and mem.dtcs == mem.rx_dtcs:
            mem.tmode = "DTCS"
        elif rxmode or txmode:
            mem.tmode = "Cross"
            mem.cross_mode = "%s->%s" % (txmode, rxmode)

        mem.dtcs_polarity = "".join(dtcs_pol)

        mem.mode = _mem.wide and "FM" or "NFM"

        mem.power = self.POWER_LEVELS[_mem.power]

        # Extra
        mem.extra = RadioSettingGroup("extra", "Extra")

        if _mem.recvmode == 0xFF:
            val = 0x00
        else:
            val = _mem.recvmode
        recvmode = RadioSetting("recvmode", "Receiving mode",
                                RadioSettingValueList(LIST_RECVMODE,
                                                      LIST_RECVMODE[val]))
        mem.extra.append(recvmode)

        if _mem.botsignal == 0xFF:
            val = 0x00
        else:
            val = _mem.botsignal
        botsignal = RadioSetting("botsignal", "Launch signaling",
                                 RadioSettingValueList(LIST_SIGNAL,
                                                       LIST_SIGNAL[val]))
        mem.extra.append(botsignal)

        if _mem.eotsignal == 0xFF:
            val = 0x00
        else:
            val = _mem.eotsignal

        rx = RadioSettingValueList(LIST_SIGNAL, LIST_SIGNAL[val])
        eotsignal = RadioSetting("eotsignal", "Transmit end signaling", rx)
        mem.extra.append(eotsignal)

        rx = RadioSettingValueBoolean(bool(_mem.compander))
        compander = RadioSetting("compander", "Compander", rx)
        mem.extra.append(compander)

        rx = RadioSettingValueBoolean(bool(_mem.scrambler))
        scrambler = RadioSetting("scrambler", "Scrambler", rx)
        mem.extra.append(scrambler)

        return mem

    def set_memory(self, mem):
        _mem = self._memory_obj()[mem.number - 1]

        if mem.empty:
            _mem.set_raw("\xff" * 24)
            _mem.namelen = 0
            return

        _mem.set_raw("\xFF" * 15 + "\x00\x00" + "\xFF" * 7)

        _mem.rxfreq = mem.freq / 10
        if mem.duplex == "off":
            _mem.txfreq = 0xFFFFFFFF
        elif mem.duplex == "split":
            _mem.txfreq = mem.offset / 10
        elif mem.duplex == "+":
            _mem.txfreq = (mem.freq + mem.offset) / 10
        elif mem.duplex == "-":
            _mem.txfreq = (mem.freq - mem.offset) / 10
        else:
            _mem.txfreq = mem.freq / 10

        _mem.namelen = len(mem.name.rstrip())
        _namelength = self.get_features().valid_name_length
        for i in range(_namelength):
            try:
                _mem.name[i] = ord(mem.name[i])
            except IndexError:
                _mem.name[i] = 0xFF

        rxmode = ""
        txmode = ""

        if mem.tmode == "Tone":
            txmode = "Tone"
        elif mem.tmode == "TSQL":
            rxmode = "Tone"
            txmode = "TSQL"
        elif mem.tmode == "DTCS":
            rxmode = "DTCSSQL"
            txmode = "DTCS"
        elif mem.tmode == "Cross":
            txmode, rxmode = mem.cross_mode.split("->", 1)

        if rxmode == "":
            _mem.rxdtcs_pol = 1
            _mem.is_rxdigtone = 1
            _mem.rxtone = 0x3FFF
        elif rxmode == "Tone":
            _mem.rxdtcs_pol = 0
            _mem.is_rxdigtone = 0
            _mem.rxtone = int(mem.ctone * 10)
        elif rxmode == "DTCSSQL":
            _mem.rxdtcs_pol = 1 if mem.dtcs_polarity[1] == "R" else 0
            _mem.is_rxdigtone = 1
            _mem.rxtone = self._set_dcs(mem.dtcs)
        elif rxmode == "DTCS":
            _mem.rxdtcs_pol = 1 if mem.dtcs_polarity[1] == "R" else 0
            _mem.is_rxdigtone = 1
            _mem.rxtone = self._set_dcs(mem.rx_dtcs)

        if txmode == "":
            _mem.txdtcs_pol = 1
            _mem.is_txdigtone = 1
            _mem.txtone = 0x3FFF
        elif txmode == "Tone":
            _mem.txdtcs_pol = 0
            _mem.is_txdigtone = 0
            _mem.txtone = int(mem.rtone * 10)
        elif txmode == "TSQL":
            _mem.txdtcs_pol = 0
            _mem.is_txdigtone = 0
            _mem.txtone = int(mem.ctone * 10)
        elif txmode == "DTCS":
            _mem.txdtcs_pol = 1 if mem.dtcs_polarity[0] == "R" else 0
            _mem.is_txdigtone = 1
            _mem.txtone = self._set_dcs(mem.dtcs)

        _mem.wide = self.MODES.index(mem.mode)
        _mem.power = mem.power == self.POWER_LEVELS[1]

        # Extra settings
        for setting in mem.extra:
            setattr(_mem, setting.get_name(), setting.value)

    def get_settings(self):
        """Translate the bit in the mem_struct into settings in the UI"""
        # Define mem struct write-back shortcuts
        _sets = self._memobj.settings
        _vfoa = self._memobj.upper.vfoa
        _vfob = self._memobj.lower.vfob
        _lims = self._memobj.hello_lims
        _codes = self._memobj.codes
        _dtmf = self._memobj.dtmf_tab

        basic = RadioSettingGroup("basic", "Basic Settings")
        a_band = RadioSettingGroup("a_band", "VFO A-Upper Settings")
        b_band = RadioSettingGroup("b_band", "VFO B-Lower Settings")
        codes = RadioSettingGroup("codes", "Codes & DTMF Groups")
        lims = RadioSettingGroup("lims", "PowerOn & Freq Limits")
        group = RadioSettings(basic, a_band, b_band, lims, codes)

        # Basic Settings
        bnd_mode = RadioSetting("settings.init_bank", "TDR Band Default",
                                RadioSettingValueList(LIST_TDR_DEF,
                                                      LIST_TDR_DEF[
                                                          _sets.init_bank]))
        basic.append(bnd_mode)

        # BJ-318 set by vfo, skip
        if self.MODEL != "BJ-318":
            rs = RadioSettingValueInteger(0, 20, _sets.volume)
            volume = RadioSetting("settings.volume", "Volume", rs)
            basic.append(volume)

        # BJ-318 set by vfo, skip
        if self.MODEL != "BJ-318":
            vala = _vfoa.bpower        # 2bits values 0,1,2= Low, Mid, High
            rx = RadioSettingValueList(LIST_BPOWER, LIST_BPOWER[vala])
            powera = RadioSetting("upper.vfoa.bpower", "Power (Upper)", rx)
            basic.append(powera)

        # BJ-318 set by vfo, skip
        if self.MODEL != "BJ-318":
            valb = _vfob.bpower
            rx = RadioSettingValueList(LIST_BPOWER, LIST_BPOWER[valb])
            powerb = RadioSetting("lower.vfob.bpower", "Power (Lower)", rx)
            basic.append(powerb)

        def my_word2raw(setting, obj, atrb, mlt=10):
            """Callback function to convert UI floating value to u16 int"""
            if str(setting.value) == "Off":
                frq = 0x0FFFF
            else:
                frq = int(float(str(setting.value)) * float(mlt))
            if frq == 0:
                frq = 0xFFFF
            setattr(obj, atrb, frq)
            return

        def my_adjraw(setting, obj, atrb, fix):
            """Callback: add or subtract fix from value."""
            vx = int(str(setting.value))
            value = vx + int(fix)
            if value < 0:
                value = 0
            if atrb == "frq_chn_mode" and int(str(setting.value)) == 2:
                value = vx * 2         # Special handling for frq_chn_mode
            setattr(obj, atrb, value)
            return

        def my_dbl2raw(setting, obj, atrb, flg=1):
            """Callback: convert from freq 146.7600 to 14760000 U32."""
            value = chirp_common.parse_freq(str(setting.value)) / 10
            # flg=1 means 0 becomes ff, else leave as possible 0
            if flg == 1 and value == 0:
                value = 0xFFFFFFFF
            setattr(obj, atrb, value)
            return

        def my_val_list(setting, obj, atrb):
            """Callback:from ValueList with non-sequential, actual values."""
            value = int(str(setting.value))            # Get the integer value
            if atrb == "tot":
                value = int(value / 30)    # 30 second increments
            setattr(obj, atrb, value)
            return

        def my_spcl(setting, obj, atrb):
            """Callback: Special handling based on atrb."""
            if atrb == "frq_chn_mode":
                idx = LIST_VFOMODE.index(str(setting.value))  # Returns 0 or 1
                value = idx * 2            # Set bit 1
            setattr(obj, atrb, value)
            return

        def my_tone_strn(obj, is_atr, pol_atr, tone_atr):
            """Generate the CTCS/DCS tone code string."""
            vx = int(getattr(obj, tone_atr))
            if vx == 16383 or vx == 0:
                return "Off"                 # 16383 is all bits set
            if getattr(obj, is_atr) == 0:             # Simple CTCSS code
                tstr = str(vx / 10.0)
            else:        # DCS
                if getattr(obj, pol_atr) == 0:
                    tstr = "D{:03x}R".format(vx)
                else:
                    tstr = "D{:03x}N".format(vx)
            return tstr

        def my_set_tone(setting, obj, is_atr, pol_atr, tone_atr):
            """Callback- create the tone setting from string code."""
            sx = str(setting.value)        # '131.8'  or 'D231N' or 'Off'
            if sx == "Off":
                isx = 1
                polx = 1
                tonx = 0x3FFF
            elif sx[0] == "D":         # DCS
                isx = 1
                if sx[4] == "N":
                    polx = 1
                else:
                    polx = 0
                tonx = int(sx[1:4], 16)
            else:                                     # CTCSS
                isx = 0
                polx = 0
                tonx = int(float(sx) * 10.0)
            setattr(obj, is_atr, isx)
            setattr(obj, pol_atr, polx)
            setattr(obj, tone_atr, tonx)
            return

        # not used by BJ-318 skip
        if self.MODEL != "BJ-318":
            val = _sets.fm_freq / 10.0
            if val == 0:
                val = 88.9            # 0 is not valid
            rx = RadioSettingValueFloat(65, 108.0, val, 0.1, 1)
            rs = RadioSetting("settings.fm_freq",
                              "FM Broadcast Freq (MHz)", rx)
            rs.set_apply_callback(my_word2raw, _sets, "fm_freq")
            basic.append(rs)

        # BJ-318 fm frequency
        if self.MODEL == "BJ-318":
            val = _lims.fm_318 / 10.0
            if val == 0:
                val = 88.9            # 0 is not valid
            rx = RadioSettingValueFloat(65, 108.0, val, 0.1, 1)
            rs = RadioSetting("hello_lims.fm_318",
                              "FM Broadcast Freq (MHz)", rx)
            rs.set_apply_callback(my_word2raw, _lims, "fm_318")
            basic.append(rs)

        # not used in BJ-318, skip
        if self.MODEL != "BJ-318":
            rs = RadioSettingValueList(LIST_COLOR,
                                       LIST_COLOR[_sets.wtled])
            wtled = RadioSetting("settings.wtled", "Standby LED Color", rs)
            basic.append(wtled)

        # not used in BJ-318, skip
        if self.MODEL != "BJ-318":
            rs = RadioSettingValueList(LIST_COLOR,
                                       LIST_COLOR[_sets.rxled])
            rxled = RadioSetting("settings.rxled", "RX LED Color", rs)
            basic.append(rxled)

        # not used in BJ-318, skip
        if self.MODEL != "BJ-318":
            rs = RadioSettingValueList(LIST_COLOR,
                                       LIST_COLOR[_sets.txled])
            txled = RadioSetting("settings.txled", "TX LED Color", rs)
            basic.append(txled)

        ledsw = RadioSetting("settings.ledsw", "Back light mode",
                             RadioSettingValueList(LIST_LEDSW, LIST_LEDSW[
                                 _sets.ledsw]))
        basic.append(ledsw)

        beep = RadioSetting("settings.beep", "Beep",
                            RadioSettingValueBoolean(bool(_sets.beep)))
        basic.append(beep)

        ring = RadioSetting("settings.ring", "Ring",
                            RadioSettingValueList(LIST_RING, LIST_RING[
                                _sets.ring]))
        basic.append(ring)

        bcl = RadioSetting("settings.bcl", "Busy channel lockout",
                           RadioSettingValueBoolean(bool(_sets.bcl)))
        basic.append(bcl)

        # squelch for non-BJ-318 models
        # BJ-318 squelch set by VFO basis
        if self.MODEL != "BJ-318":
            if _vfoa.sql == 0xFF:
                val = 0x04
            else:
                val = _vfoa.sql
            sqla = RadioSetting("upper.vfoa.sql", "Squelch (Upper)",
                                RadioSettingValueInteger(0, 9, val))
            basic.append(sqla)

            if _vfob.sql == 0xFF:
                val = 0x04
            else:
                val = _vfob.sql
            sqlb = RadioSetting("lower.vfob.sql", "Squelch (Lower)",
                                RadioSettingValueInteger(0, 9, val))
            basic.append(sqlb)

        tmp = str(int(_sets.tot) * 30)     # 30 sec step counter
        rs = RadioSetting("settings.tot", "Transmit Timeout (Secs)",
                          RadioSettingValueList(LIST_TIMEOUT, tmp))
        rs.set_apply_callback(my_val_list, _sets, "tot")
        basic.append(rs)

        tmp = str(int(_sets.sig_freq))
        rs = RadioSetting("settings.sig_freq", "Single Signaling Tone (Hz)",
                          RadioSettingValueList(LIST_SSF, tmp))
        rs.set_apply_callback(my_val_list, _sets, "sig_freq")
        basic.append(rs)

        tmp = str(int(_sets.dtmf_txms))
        rs = RadioSetting("settings.dtmf_txms", "DTMF Tx Duration (mSecs)",
                          RadioSettingValueList(LIST_DTMFTX, tmp))
        rs.set_apply_callback(my_val_list, _sets, "dtmf_txms")
        basic.append(rs)

        rs = RadioSetting("settings.rptr_mode", "Repeater Mode",
                          RadioSettingValueBoolean(bool(_sets.rptr_mode)))
        basic.append(rs)

        # UPPER BAND SETTINGS

        # Freq Mode, convert bit 1 state to index pointer
        val = _vfoa.frq_chn_mode // 2

        rx = RadioSettingValueList(LIST_VFOMODE, LIST_VFOMODE[val])
        rs = RadioSetting("upper.vfoa.frq_chn_mode", "Default Mode", rx)
        rs.set_apply_callback(my_spcl, _vfoa, "frq_chn_mode")
        a_band.append(rs)

        val = _vfoa.chan_num + 1                  # Add 1 for 1-128 displayed
        rs = RadioSetting("upper.vfoa.chan_num", "Initial Chan",
                          RadioSettingValueInteger(1, 128, val))
        rs.set_apply_callback(my_adjraw, _vfoa, "chan_num", -1)
        a_band.append(rs)

        val = _vfoa.rxfreq / 100000.0
        if (val < 136.0 or val > 176.0):
            val = 146.520            # 2m calling
        rs = RadioSetting("upper.vfoa.rxfreq ", "Default Recv Freq (MHz)",
                          RadioSettingValueFloat(136.0, 176.0, val, 0.001, 5))
        rs.set_apply_callback(my_dbl2raw, _vfoa, "rxfreq")
        a_band.append(rs)

        tmp = my_tone_strn(_vfoa, "is_rxdigtone", "rxdtcs_pol", "rx_tone")
        rs = RadioSetting("rx_tone", "Default Recv CTCSS (Hz)",
                          RadioSettingValueList(LIST_CTCSS, tmp))
        rs.set_apply_callback(my_set_tone, _vfoa, "is_rxdigtone",
                              "rxdtcs_pol", "rx_tone")
        a_band.append(rs)

        rx = RadioSettingValueList(LIST_RECVMODE,
                                   LIST_RECVMODE[_vfoa.rx_mode])
        rs = RadioSetting("upper.vfoa.rx_mode", "Default Recv Mode", rx)
        a_band.append(rs)

        tmp = my_tone_strn(_vfoa, "is_txdigtone", "txdtcs_pol", "tx_tone")
        rs = RadioSetting("tx_tone", "Default Xmit CTCSS (Hz)",
                          RadioSettingValueList(LIST_CTCSS, tmp))
        rs.set_apply_callback(my_set_tone, _vfoa, "is_txdigtone",
                              "txdtcs_pol", "tx_tone")
        a_band.append(rs)

        rs = RadioSetting("upper.vfoa.launch_sig", "Launch Signaling",
                          RadioSettingValueList(LIST_SIGNAL,
                                                LIST_SIGNAL[_vfoa.launch_sig]))
        a_band.append(rs)

        rx = RadioSettingValueList(LIST_SIGNAL, LIST_SIGNAL[_vfoa.tx_end_sig])
        rs = RadioSetting("upper.vfoa.tx_end_sig", "Xmit End Signaling", rx)
        a_band.append(rs)

        rx = RadioSettingValueList(LIST_BW, LIST_BW[_vfoa.fm_bw])
        rs = RadioSetting("upper.vfoa.fm_bw", "Wide/Narrow Band", rx)
        a_band.append(rs)

        rx = RadioSettingValueBoolean(bool(_vfoa.cmp_nder))
        rs = RadioSetting("upper.vfoa.cmp_nder", "Compander", rx)
        a_band.append(rs)

        rs = RadioSetting("upper.vfoa.scrm_blr", "Scrambler",
                          RadioSettingValueBoolean(bool(_vfoa.scrm_blr)))
        a_band.append(rs)

        rx = RadioSettingValueList(LIST_SHIFT, LIST_SHIFT[_vfoa.shift])
        rs = RadioSetting("upper.vfoa.shift", "Xmit Shift", rx)
        a_band.append(rs)

        val = _vfoa.offset / 100000.0
        rs = RadioSetting("upper.vfoa.offset", "Xmit Offset (MHz)",
                          RadioSettingValueFloat(0, 100.0, val, 0.001, 3))
        # Allow zero value
        rs.set_apply_callback(my_dbl2raw, _vfoa, "offset", 0)
        a_band.append(rs)

        tmp = str(_vfoa.step / 100.0)
        rs = RadioSetting("step", "Freq step (kHz)",
                          RadioSettingValueList(LIST_STEPS, tmp))
        rs.set_apply_callback(my_word2raw, _vfoa, "step", 100)
        a_band.append(rs)

        # BJ-318 upper band squelch
        if self.MODEL == "BJ-318":
            if _vfoa.sql == 0xFF:
                valq = 0x04    # setting default squelch to 04
            else:
                valq = _vfoa.sql
                sqla = RadioSetting("upper.vfoa.sql", "Squelch",
                                    RadioSettingValueInteger(0, 9, valq))
            a_band.append(sqla)

        # BJ-318 upper band volume
        if self.MODEL == "BJ-318":
            bvolume_u = RadioSetting("upper.vfoa.bvol",
                                     "Volume", RadioSettingValueInteger(
                                         0, 15, _vfoa.bvol))
            a_band.append(bvolume_u)

        # BJ-318 Upper Screen Color
        if self.MODEL == "BJ-318":
            rs_u = RadioSettingValueList(LIST_COLOR318,
                                         LIST_COLOR318[_lims.up_scr_color])
            up_scr_color = RadioSetting("hello_lims.up_scr_color",
                                        "Screen Color", rs_u)
            a_band.append(up_scr_color)

        # BJ-318 Upper band power
        if self.MODEL == "BJ-318":
            val_b = _vfoa.bpower        # 2bits values 0,1,2= Low, Mid, High
            rx = RadioSettingValueList(LIST_BPOWER, LIST_BPOWER[val_b])
            power_u = RadioSetting("upper.vfoa.bpower", "Power", rx)
            a_band.append(power_u)

        # LOWER BAND SETTINGS

        val = _vfob.frq_chn_mode // 2
        rx = RadioSettingValueList(LIST_VFOMODE, LIST_VFOMODE[val])
        rs = RadioSetting("lower.vfob.frq_chn_mode", "Default Mode", rx)
        rs.set_apply_callback(my_spcl, _vfob, "frq_chn_mode")
        b_band.append(rs)

        val = _vfob.chan_num + 1
        rs = RadioSetting("lower.vfob.chan_num", "Initial Chan",
                          RadioSettingValueInteger(1, 128, val))
        rs.set_apply_callback(my_adjraw, _vfob, "chan_num", -1)
        b_band.append(rs)

        val = _vfob.rxfreq / 100000.0
        if (val < 400.0 or val > 490.0):
            val = 446.0          # UHF calling
        rs = RadioSetting("lower.vfob.rxfreq ", "Default Recv Freq (MHz)",
                          RadioSettingValueFloat(400.0, 490.0, val, 0.001, 5))
        rs.set_apply_callback(my_dbl2raw, _vfob, "rxfreq")
        b_band.append(rs)

        tmp = my_tone_strn(_vfob, "is_rxdigtone", "rxdtcs_pol", "rx_tone")
        rs = RadioSetting("rx_tone", "Default Recv CTCSS (Hz)",
                          RadioSettingValueList(LIST_CTCSS, tmp))
        rs.set_apply_callback(my_set_tone, _vfob, "is_rxdigtone",
                              "rxdtcs_pol", "rx_tone")
        b_band.append(rs)

        rx = RadioSettingValueList(LIST_RECVMODE, LIST_RECVMODE[_vfob.rx_mode])
        rs = RadioSetting("lower.vfob.rx_mode", "Default Recv Mode", rx)
        b_band.append(rs)

        tmp = my_tone_strn(_vfob, "is_txdigtone", "txdtcs_pol", "tx_tone")
        rs = RadioSetting("tx_tone", "Default Xmit CTCSS (Hz)",
                          RadioSettingValueList(LIST_CTCSS, tmp))
        rs.set_apply_callback(my_set_tone, _vfob, "is_txdigtone",
                              "txdtcs_pol", "tx_tone")
        b_band.append(rs)

        rx = RadioSettingValueList(LIST_SIGNAL, LIST_SIGNAL[_vfob.launch_sig])
        rs = RadioSetting("lower.vfob.launch_sig", "Launch Signaling", rx)
        b_band.append(rs)

        rx = RadioSettingValueList(LIST_SIGNAL, LIST_SIGNAL[_vfob.tx_end_sig])
        rs = RadioSetting("lower.vfob.tx_end_sig", "Xmit End Signaling", rx)
        b_band.append(rs)

        rx = RadioSettingValueList(LIST_BW, LIST_BW[_vfob.fm_bw])
        rs = RadioSetting("lower.vfob.fm_bw", "Wide/Narrow Band", rx)
        b_band.append(rs)

        rs = RadioSetting("lower.vfob.cmp_nder", "Compander",
                          RadioSettingValueBoolean(bool(_vfob.cmp_nder)))
        b_band.append(rs)

        rs = RadioSetting("lower.vfob.scrm_blr", "Scrambler",
                          RadioSettingValueBoolean(bool(_vfob.scrm_blr)))
        b_band.append(rs)

        rx = RadioSettingValueList(LIST_SHIFT, LIST_SHIFT[_vfob.shift])
        rs = RadioSetting("lower.vfob.shift", "Xmit Shift", rx)
        b_band.append(rs)

        val = _vfob.offset / 100000.0
        rs = RadioSetting("lower.vfob.offset", "Xmit Offset (MHz)",
                          RadioSettingValueFloat(0, 100.0, val, 0.001, 3))
        rs.set_apply_callback(my_dbl2raw, _vfob, "offset", 0)
        b_band.append(rs)

        tmp = str(_vfob.step / 100.0)
        rs = RadioSetting("step", "Freq step (kHz)",
                          RadioSettingValueList(LIST_STEPS, tmp))
        rs.set_apply_callback(my_word2raw, _vfob, "step", 100)
        b_band.append(rs)

        # BJ-318 lower band squelch
        if self.MODEL == "BJ-318":
            if _vfob.sql == 0xFF:
                val_l = 0x04    # setting default squelch to 04
            else:
                val_l = _vfob.sql
                sql_b = RadioSetting("lower.vfob.sql", "Squelch",
                                     RadioSettingValueInteger(0, 9, val_l))
                b_band.append(sql_b)

        # BJ-318 lower band volume
        if self.MODEL == "BJ-318":
            bvolume_l = RadioSetting("lower.vfob.bvol",
                                     "Volume", RadioSettingValueInteger(
                                         0, 15, _vfob.bvol))
            b_band.append(bvolume_l)

        # BJ-318 Lower Screen Color
        if self.MODEL == "BJ-318":
            rs_l = RadioSettingValueList(LIST_COLOR318,
                                         LIST_COLOR318[_lims.dn_scr_color])
            dn_scr_color = RadioSetting("hello_lims.dn_scr_color",
                                        "Screen Color", rs_l)
            b_band.append(dn_scr_color)

        # BJ-318 lower band power
        if self.MODEL == "BJ-318":
            val_l = _vfoa.bpower        # 2bits values 0,1,2= Low, Mid, High
            rx = RadioSettingValueList(LIST_BPOWER, LIST_BPOWER[val_l])
            powera = RadioSetting("lower.vfob.bpower", "Power", rx)
            b_band.append(powera)

        # PowerOn & Freq Limits Settings
        def chars2str(cary, knt):
            """Convert raw memory char array to a string: NOT a callback."""
            stx = ""
            for char in cary[0:knt]:
                stx += chr(int(char))
            return stx

        def my_str2ary(setting, obj, atrba, atrbc):
            """Callback: convert 7-char string to char array with count."""
            ary = ""
            knt = 7
            for j in range(6, -1, -1):       # Strip trailing spaces
                if str(setting.value)[j] == "" or str(setting.value)[j] == " ":
                    knt = knt - 1
                else:
                    break
            for j in range(0, 7, 1):
                if j < knt:
                    ary += str(setting.value)[j]
                else:
                    ary += chr(0xFF)
            setattr(obj, atrba, ary)
            setattr(obj, atrbc, knt)
            return

        # not used in BJ-318 startup screen
        if self.MODEL != "BJ-318":
            tmp = chars2str(_lims.hello1, _lims.hello1_cnt)
            rs = RadioSetting("hello_lims.hello1", "Power-On Message 1",
                              RadioSettingValueString(0, 7, tmp))
            rs.set_apply_callback(my_str2ary, _lims, "hello1", "hello1_cnt")
            lims.append(rs)

        # not used in BJ-318 startup screen
        if self.MODEL != "BJ-318":
            tmp = chars2str(_lims.hello2, _lims.hello2_cnt)
            rs = RadioSetting("hello_lims.hello2", "Power-On Message 2",
                              RadioSettingValueString(0, 7, tmp))
            rs.set_apply_callback(my_str2ary, _lims, "hello2", "hello2_cnt")
            lims.append(rs)

        # VALID_BANDS = [(136000000, 176000000),400000000, 490000000)]

        lval = _lims.vhf_low / 100000.0
        uval = _lims.vhf_high / 100000.0
        if lval >= uval:
            lval = 144.0
            uval = 158.0

        rs = RadioSetting("hello_lims.vhf_low", "Lower VHF Band Limit (MHz)",
                          RadioSettingValueFloat(136.0, 176.0, lval, 0.001, 3))
        rs.set_apply_callback(my_dbl2raw, _lims, "vhf_low")
        lims.append(rs)

        rs = RadioSetting("hello_lims.vhf_high", "Upper VHF Band Limit (MHz)",
                          RadioSettingValueFloat(136.0, 176.0, uval, 0.001, 3))
        rs.set_apply_callback(my_dbl2raw, _lims, "vhf_high")
        lims.append(rs)

        lval = _lims.uhf_low / 100000.0
        uval = _lims.uhf_high / 100000.0
        if lval >= uval:
            lval = 420.0
            uval = 470.0

        rs = RadioSetting("hello_lims.uhf_low", "Lower UHF Band Limit (MHz)",
                          RadioSettingValueFloat(400.0, 490.0, lval, 0.001, 3))
        rs.set_apply_callback(my_dbl2raw, _lims, "uhf_low")
        lims.append(rs)

        rs = RadioSetting("hello_lims.uhf_high", "Upper UHF Band Limit (MHz)",
                          RadioSettingValueFloat(400.0, 490.0, uval, 0.001, 3))
        rs.set_apply_callback(my_dbl2raw, _lims, "uhf_high")
        lims.append(rs)

        # Codes and DTMF Groups Settings

        def make_dtmf(ary, knt):
            """Generate the DTMF code 1-8, NOT a callback."""
            tmp = ""
            if knt > 0 and knt != 0xff:
                for val in ary[0:knt]:
                    if val > 0 and val <= 9:
                        tmp += chr(val + 48)
                    elif val == 0x0a:
                        tmp += "0"
                    elif val == 0x0d:
                        tmp += "A"
                    elif val == 0x0e:
                        tmp += "B"
                    elif val == 0x0f:
                        tmp += "C"
                    elif val == 0x00:
                        tmp += "D"
                    elif val == 0x0b:
                        tmp += "*"
                    elif val == 0x0c:
                        tmp += "#"
                    else:
                        msg = ("Invalid Character. Must be: 0-9,A,B,C,D,*,#")
                        raise InvalidValueError(msg)
            return tmp

        def my_dtmf2raw(setting, obj, atrba, atrbc, syz=7):
            """Callback: DTMF Code; sends 5 or 7-byte string."""
            draw = []
            knt = syz
            for j in range(syz - 1, -1, -1):       # Strip trailing spaces
                if str(setting.value)[j] == "" or str(setting.value)[j] == " ":
                    knt = knt - 1
                else:
                    break
            for j in range(0, syz):
                bx = str(setting.value)[j]
                obx = ord(bx)
                dig = 0x0ff
                if j < knt and knt > 0:      # (Else) is pads
                    if bx == "0":
                        dig = 0x0a
                    elif bx == "A":
                        dig = 0x0d
                    elif bx == "B":
                        dig = 0x0e
                    elif bx == "C":
                        dig = 0x0f
                    elif bx == "D":
                        dig = 0x00
                    elif bx == "*":
                        dig = 0x0b
                    elif bx == "#":
                        dig = 0x0c
                    elif obx >= 49 and obx <= 57:
                        dig = obx - 48
                    else:
                        msg = ("Must be: 0-9,A,B,C,D,*,#")
                        raise InvalidValueError(msg)
                    # - End if/elif/else for bx
                # - End if J<=knt
                draw.append(dig)         # Generate string of bytes
            # - End for j
            setattr(obj, atrba, draw)
            setattr(obj, atrbc, knt)
            return

        tmp = make_dtmf(_codes.native_id_code, _codes.native_id_cnt)
        rs = RadioSetting("codes.native_id_code", "Native ID Code",
                          RadioSettingValueString(0, 7, tmp))
        rs.set_apply_callback(my_dtmf2raw, _codes, "native_id_code",
                              "native_id_cnt", 7)
        codes.append(rs)

        tmp = make_dtmf(_codes.master_id_code, _codes.master_id_cnt)
        rs = RadioSetting("codes.master_id_code", "Master Control ID Code",
                          RadioSettingValueString(0, 7, tmp))
        rs.set_apply_callback(my_dtmf2raw, _codes, "master_id_code",
                              "master_id_cnt", 7)
        codes.append(rs)

        tmp = make_dtmf(_codes.alarm_code, _codes.alarm_cnt)
        rs = RadioSetting("codes.alarm_code", "Alarm Code",
                          RadioSettingValueString(0, 5, tmp))
        rs.set_apply_callback(my_dtmf2raw, _codes, "alarm_code",
                              "alarm_cnt", 5)
        codes.append(rs)

        tmp = make_dtmf(_codes.id_disp_code, _codes.id_disp_cnt)
        rs = RadioSetting("codes.id_disp_code", "Identify Display Code",
                          RadioSettingValueString(0, 5, tmp))
        rs.set_apply_callback(my_dtmf2raw, _codes, "id_disp_code",
                              "id_disp_cnt", 5)
        codes.append(rs)

        tmp = make_dtmf(_codes.revive_code, _codes.revive_cnt)
        rs = RadioSetting("codes.revive_code", "Revive Code",
                          RadioSettingValueString(0, 5, tmp))
        rs.set_apply_callback(my_dtmf2raw, _codes, "revive_code",
                              "revive_cnt", 5)
        codes.append(rs)

        tmp = make_dtmf(_codes.stun_code, _codes.stun_cnt)
        rs = RadioSetting("codes.stun_code", "Remote Stun Code",
                          RadioSettingValueString(0, 5, tmp))
        rs.set_apply_callback(my_dtmf2raw,  _codes, "stun_code",
                              "stun_cnt", 5)
        codes.append(rs)

        tmp = make_dtmf(_codes.kill_code, _codes.kill_cnt)
        rs = RadioSetting("codes.kill_code", "Remote KILL Code",
                          RadioSettingValueString(0, 5, tmp))
        rs.set_apply_callback(my_dtmf2raw, _codes, "kill_code",
                              "kill_cnt", 5)
        codes.append(rs)

        tmp = make_dtmf(_codes.monitor_code, _codes.monitor_cnt)
        rs = RadioSetting("codes.monitor_code", "Monitor Code",
                          RadioSettingValueString(0, 5, tmp))
        rs.set_apply_callback(my_dtmf2raw, _codes, "monitor_code",
                              "monitor_cnt", 5)
        codes.append(rs)

        val = _codes.state_now
        if val > 2:
            val = 0

        rx = RadioSettingValueList(LIST_STATE, LIST_STATE[val])
        rs = RadioSetting("codes.state_now", "Current State", rx)
        codes.append(rs)

        dtm = make_dtmf(_dtmf.dtmf1, _dtmf.dtmf1_cnt)
        rs = RadioSetting("dtmf_tab.dtmf1", "DTMF1 String",
                          RadioSettingValueString(0, 7, dtm))
        rs.set_apply_callback(my_dtmf2raw, _dtmf, "dtmf1", "dtmf1_cnt")
        codes.append(rs)

        dtm = make_dtmf(_dtmf.dtmf2, _dtmf.dtmf2_cnt)
        rs = RadioSetting("dtmf_tab.dtmf2", "DTMF2 String",
                          RadioSettingValueString(0, 7, dtm))
        rs.set_apply_callback(my_dtmf2raw, _dtmf, "dtmf2", "dtmf2_cnt")
        codes.append(rs)

        dtm = make_dtmf(_dtmf.dtmf3, _dtmf.dtmf3_cnt)
        rs = RadioSetting("dtmf_tab.dtmf3", "DTMF3 String",
                          RadioSettingValueString(0, 7, dtm))
        rs.set_apply_callback(my_dtmf2raw, _dtmf, "dtmf3", "dtmf3_cnt")
        codes.append(rs)

        dtm = make_dtmf(_dtmf.dtmf4, _dtmf.dtmf4_cnt)
        rs = RadioSetting("dtmf_tab.dtmf4", "DTMF4 String",
                          RadioSettingValueString(0, 7, dtm))
        rs.set_apply_callback(my_dtmf2raw, _dtmf, "dtmf4", "dtmf4_cnt")
        codes.append(rs)

        dtm = make_dtmf(_dtmf.dtmf5, _dtmf.dtmf5_cnt)
        rs = RadioSetting("dtmf_tab.dtmf5", "DTMF5 String",
                          RadioSettingValueString(0, 7, dtm))
        rs.set_apply_callback(my_dtmf2raw, _dtmf, "dtmf5", "dtmf5_cnt")
        codes.append(rs)

        dtm = make_dtmf(_dtmf.dtmf6, _dtmf.dtmf6_cnt)
        rs = RadioSetting("dtmf_tab.dtmf6", "DTMF6 String",
                          RadioSettingValueString(0, 7, dtm))
        rs.set_apply_callback(my_dtmf2raw, _dtmf, "dtmf6", "dtmf6_cnt")
        codes.append(rs)

        dtm = make_dtmf(_dtmf.dtmf7, _dtmf.dtmf7_cnt)
        rs = RadioSetting("dtmf_tab.dtmf7", "DTMF7 String",
                          RadioSettingValueString(0, 7, dtm))
        rs.set_apply_callback(my_dtmf2raw, _dtmf, "dtmf7", "dtmf7_cnt")
        codes.append(rs)

        dtm = make_dtmf(_dtmf.dtmf8, _dtmf.dtmf8_cnt)
        rs = RadioSetting("dtmf_tab.dtmf8", "DTMF8 String",
                          RadioSettingValueString(0, 7, dtm))
        rs.set_apply_callback(my_dtmf2raw, _dtmf, "dtmf8", "dtmf8_cnt")
        codes.append(rs)

        return group       # END get_settings()

    def set_settings(self, settings):
        _settings = self._memobj.settings
        _mem = self._memobj
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            else:
                try:
                    name = element.get_name()
                    if "." in name:
                        bits = name.split(".")
                        obj = self._memobj
                        for bit in bits[:-1]:
                            if "/" in bit:
                                bit, index = bit.split("/", 1)
                                index = int(index)
                                obj = getattr(obj, bit)[index]
                            else:
                                obj = getattr(obj, bit)
                        setting = bits[-1]
                    else:
                        obj = _settings
                        setting = element.get_name()

                    if element.has_apply_callback():
                        LOG.debug("Using apply callback")
                        element.run_apply_callback()
                    elif element.value.get_mutable():
                        LOG.debug("Setting %s = %s" % (setting, element.value))
                        setattr(obj, setting, element.value)
                except Exception:
                    LOG.debug(element.get_name())
                    raise

    @classmethod
    def match_model(cls, filedata, filename):
        match_size = False
        match_model = False

        # Testing the file data size
        if len(filedata) == MEM_SIZE + 8:
            match_size = True

        # Testing the firmware model fingerprint
        match_model = model_match(cls, filedata)

        if match_size and match_model:
            return True
        else:
            return False


class LT725UVUpper(LT725UV):
    VARIANT = "Upper"
    _vfo = "upper"


class LT725UVLower(LT725UV):
    VARIANT = "Lower"
    _vfo = "lower"


class Zastone(chirp_common.Alias):
    """Declare BJ-218 alias for Zastone BJ-218."""
    VENDOR = "Zastone"
    MODEL = "BJ-218"


class Hesenate(chirp_common.Alias):
    """Declare BJ-218 alias for Hesenate BJ-218."""
    VENDOR = "Hesenate"
    MODEL = "BJ-218"


class Baojie218Upper(LT725UVUpper):
    VENDOR = "Baojie"
    MODEL = "BJ-218"


class Baojie218Lower(LT725UVLower):
    VENDOR = "Baojie"
    MODEL = "BJ-218"


@directory.register
class Baojie218(LT725UV):
    """Baojie BJ-218"""
    VENDOR = "Baojie"
    MODEL = "BJ-218"
    ALIASES = [Zastone, Hesenate, ]

    def get_sub_devices(self):
        return [Baojie218Upper(self._mmap), Baojie218Lower(self._mmap)]


class Baojie318Upper(LT725UVUpper):
    VENDOR = "Baojie"
    MODEL = "BJ-318"


class Baojie318Lower(LT725UVLower):
    VENDOR = "Baojie"
    MODEL = "BJ-318"


@directory.register
class Baojie318(LT725UV):
    """Baojie BJ-318"""
    VENDOR = "Baojie"
    MODEL = "BJ-318"

    def get_sub_devices(self):
        return [Baojie318Upper(self._mmap), Baojie318Lower(self._mmap)]
